"""
Examples of using the lazycls extension for cmd

Create a new Type[Cmd]
"""
from lazy.cmd import Cmd

aws = Cmd('aws')

buckets = aws('s3', 'ls').val

"""
>> Returns a List[str] of your buckets

['2021-12-22 13:16:17 bucket1',
'2021-11-21 16:07:00 bucket2',
'2021-12-05 21:38:56 bucket3']

"""

"""
Example usage of automatically generated Type[Cmd] based off of local executables
"""

from lazy.cmd import export

"""
Example usage of gsutil 
"""
from lazy.cmd.auto import gsutil

buckets = gsutil('ls').val

"""
>> Returns a List[str] of your buckets

['gs://bucket1',
 'gs://bucket2',
 'gs://bucket3',
 'gs://bucket4']
"""

bucket_paths = gsutil('ls', 'gs://bucket1').val

"""
>> Returns a List[str] of your bucket/path

['gs://bucket1/data/',
 'gs://bucket1/files/',
 'gs://bucket1/shared/',
 'gs://bucket1/public/']
"""


"""
Example usage of kubectl 
"""
from lazy.cmd.auto import kubectl

export(KUBECONFIG='/path/to/kubeconfig')
nodes = kubectl('get', 'nodes').as_table()

"""
>> Returns a List[Dict] of your Cluster Nodes

[{'NAME': '...ec2.internal',
  'STATUS': 'Ready',
  'ROLES': '<none>',
  'AGE': '3d15h',
  'VERSION': 'v1.21.5-eks-bc4871b'},
 {'NAME': '...ec2.internal',
  'STATUS': 'Ready',
  'ROLES': '<none>',
  'AGE': '10d',
  'VERSION': 'v1.21.5-eks-bc4871b'},
 {'NAME': '...ec2.internal',
  'STATUS': 'Ready',
  'ROLES': '<none>',
  'AGE': '3d8h',
  'VERSION': 'v1.21.5-eks-bc4871b'}]
"""

pods = kubectl('get', 'pods', '--namespace=ingress-nginx').as_table()

"""
>> Returns a List[Dict] of your Pods in namespace ingress-nginx

[{'NAME': 'ingress-nginx-controller-566d4c67fd-5dz7m',
  'READY': '1/1',
  'STATUS': 'Running',
  'RESTARTS': '0',
  'AGE': '2d23h'},
 {'NAME': 'ingress-nginx-controller-566d4c67fd-6k2jr',
  'READY': '1/1',
  'STATUS': 'Running',
  'RESTARTS': '0',
  'AGE': '10d'},
  ...
]
"""





import os
import sys
import signal
import shlex
import threading
import subprocess
import multiprocessing

from typing import Union, List
from lazy.types import classproperty
from lazy.io.pathz import get_path, PathLike, get_lazydir
from lazy.io.cachez import Cache
from .utils import *

runz_dir: PathLike = None #get_lazydir(True).joinpath('runz')
runzCache: Cache = None #  = Cache(directory = )

def get_runzcache():
    global runzCache, runz_dir
    if runzCache is None:
        runz_dir = get_lazydir(True).joinpath('runz')
        runzCache = Cache(directory = runz_dir.as_posix())
    return runzCache


class SubProcArgs:
    base = {'stdout': subprocess.PIPE, 'stderr': subprocess.STDOUT}
    win = {'creationflags': subprocess.DETACHED_PROCESS | subprocess.CREATE_NEW_PROCESS_GROUP}
    darwin = {'preexec_fn': os.setpgrp}
    linux = {}
    
    @classmethod
    def get_host_args(cls):
        if sys.platform.startswith('win'): return cls.win
        if sys.platform.startswith('darwin'): return cls.darwin
        return cls.linux

    @classmethod
    def get_background_args(cls, **kwargs):
        args = cls.get_host_args()
        args.update(cls.base)
        if kwargs: args.update(kwargs)


class RunzType(type):

    @classmethod
    def run_system(cls, cmd: str):
        """
        alias for os.system(cmd)
        """
        return os.system(cmd)

    @classmethod
    def run_cmd(cls, cmd: Union[List[str], str], shell: bool = True, raise_error: bool = True, **kwargs):
        """
        Uses subprocess.check_output(cmd, shell=shell, **kwargs)
        """
        if isinstance(cmd, str): cmd = shlex.split(cmd)
        try:
            out = subprocess.check_output(cmd, shell=shell, **kwargs)
            if isinstance(out, bytes): out = out.decode('utf8')
            return out.strip()
        except Exception as e:
            if not raise_error: return ""
            raise e
    
    @classmethod
    def sp_run(cls, cmd: Union[List[str], str], shell: bool = True, stdout = subprocess.PIPE, stderr = subprocess.STDOUT, **kwargs):
        """
        Uses subprocess.run(cmd, shell=shell, stdout=stdout, stderr=stderr, **kwargs)
        """
        if isinstance(cmd, str): cmd = shlex.split(cmd)
        return subprocess.run(cmd, shell=shell, stdout=stdout, stderr=stderr, **kwargs)
    
    @classmethod
    def sp_output(cls, cmd: Union[List[str], str], shell: bool = True, stdout = subprocess.PIPE, stderr = subprocess.STDOUT, **kwargs):
        """
        Uses subprocess.check_output(cmd, shell=shell, stdout=stdout, stderr=stderr, **kwargs)
        """
        if isinstance(cmd, str): cmd = shlex.split(cmd)
        return subprocess.check_output(cmd, shell=shell, stdout=stdout, stderr=stderr, **kwargs)
    

    @classmethod
    def sp_open(cls, cmd: Union[List[str], str], shell: bool = True, stdout = subprocess.PIPE, stderr = subprocess.STDOUT, new_process: bool = False, **kwargs):
        """
        Uses subprocess.Popen(cmd, shell=shell, stdout=stdout, stderr=stderr, **kwargs)

        if new_process, sets the args for the host system to run in an independent process.
        """
        if isinstance(cmd, str): cmd = shlex.split(cmd)
        args = SubProcArgs.get_host_args() if new_process else {}
        if kwargs: args.update(kwargs)
        return subprocess.Popen(cmd, shell=shell, stdout=stdout, stderr=stderr, **args)
    

    @classmethod
    def exec_sed(path: Union[str, PathLike], changes: List[set], new_path: Union[str, PathLike] = None, chmod: int = None) -> PathLike:
        """ Does a cleaner `sed` command 
            - path: path to file
            - changes: List of {find, replace, n_times}
            - new_path: new filepath, otherwise will write to original
            - chmod: int 
        """
        f = get_path(path)
        txt = f.read_text(encoding='utf-8')
        for c in changes:
            txt = txt.replace(str(c[0]), str(c[1]), c[2]) if len(c) == 3 else txt.replace(str(c[0]), str(c[1]))
        if new_path: f = get_path(new_path)
        f.write_text(txt, encoding='utf-8')
        if chmod: f.chmod(chmod)
        return f
    

    
    

    


import os
import time
import threading
import multiprocessing
import platform
import subprocess

from typing import Union
from lazy.utils import _get_logger

logger = _get_logger('Fuze')

def get_multiproc():
    # We can't use forkserver because we have to make sure
    # that the server inherits the per-test stdout/stderr file
    # descriptors.
    if hasattr(multiprocessing, 'get_context'): mp = multiprocessing.get_context('fork')
    else: mp = multiprocessing
    #if threading.active_count() != 1: raise RuntimeError("Multi-threaded test running is not supported")
    if threading.active_count() != 1: pass
    return mp


def exitcode(process: Union[subprocess.Popen, multiprocessing.Process]):
    if isinstance(process, subprocess.Popen): return process.poll()
    if process.is_alive(): return None
    else: return process.exitcode

def wait_for(callable, timeout=10, interval=0.1):
    '''Wait until *callable* returns something True and return it
    If *timeout* expires, return None
    '''

    waited = 0
    while True:
        ret = callable()
        if ret: return ret
        if waited > timeout: return None
        waited += interval
        time.sleep(interval)



def cleanup_fuze(mount_process: Union[subprocess.Popen, multiprocessing.Process], mnt_dir: str):
    if platform.system() == 'Darwin':
        subprocess.call(['umount', '-l', mnt_dir], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
    else:
        subprocess.call(['fusermount', '-z', '-u', mnt_dir], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

    mount_process.terminate()
    if isinstance(mount_process, subprocess.Popen):
        try: mount_process.wait(1)
        except subprocess.TimeoutExpired: mount_process.kill()
    else:
        mount_process.join(5)
        if mount_process.exitcode is None: mount_process.kill()


def umount(mnt_dir: str, mount_process: Union[subprocess.Popen, multiprocessing.Process] = None):
    if platform.system() == 'Darwin': subprocess.check_call(['umount', '-l', mnt_dir])
    else: subprocess.check_call(['fusermount', '-z', '-u', mnt_dir])
    assert not os.path.ismount(mnt_dir)
    if not mount_process: return

    if isinstance(mount_process, subprocess.Popen):
        try:
            code = mount_process.wait(5)
            if code == 0: return
            logger.error('file system process terminated with code %s' % (code,))
        except subprocess.TimeoutExpired:
            mount_process.terminate()
            try: mount_process.wait(1)
            except subprocess.TimeoutExpired: mount_process.kill()
    else:
        mount_process.join(5)
        code = mount_process.exitcode
        if code == 0: return
        elif code is None:
            mount_process.terminate()
            mount_process.join(1)
        else:
            logger.error('file system process terminated with code %s' % (code,))

    logger.error('mount process did not terminate')


from . import types
from . import base
from . import core
from . import cloud
from . import generic

from .generic import as_path, get_path, PathLike, register_pathlike_cls, get_userhome, get_lazydir

to_path = get_path

from .core import PosixFSxPath, WindowsFSxPath
from .cloud import PosixGCSPath, PosixS3Path, WindowsS3Path, WindowsGCSPath

import pathlib
PathzPath = (PosixFSxPath, WindowsFSxPath, PosixGCSPath, PosixS3Path, WindowsS3Path, WindowsGCSPath, pathlib.Path, pathlib.PosixPath, pathlib.PurePath, str)


from . import static # zero deps
from . import types # zero deps
from . import libz # zero deps
from . import serialize # zero deps
from . import utils

## Minor Deps
from . import models # Depends on types
from . import configz # depends on models, types

## Moderate Deps
from . import io # depends on serialize, configz
from . import cmd # depends on types, serialize, io

## Heavy Deps


## Bring submodules up

from .types import classproperty
from .libz import Lib
from .serialize import Serialize
from .utils.helpers import get_logger, get_cls_logger

from .models import (
    BaseCls,
    BaseModel,
    BaseLazy,
    Field,
    LazyCls,
    create_lazycls,
    clear_lazy_models,
    set_modulename,
)
from .configz import ConfigCls, CloudAuthz

ConfigTypes = configz.types
BaseTypes = types.base

from .io import get_path, PathLike, PathzPath
from .cmd import Cmd




"""
Base Types from Typing
"""

__all__ = (
    'PathLike',
    'TYPE_CHECKING',
    'List', 'Dict', 'AnyStr', 'Any',
    'Optional', 'Union', 'Tuple', 'Mapping', 'Sequence', 'TypeVar', 'Type',
    'Callable', 'Coroutine', 'Generator', 'IO', 'Iterable', 'Iterator', 'AsyncIterator',
    'cast', 'overload',
    'Final',
    'Data', 'AnyMany', 'TextMany', 'TextList',
    'DictList', 'DictMany', 'DictAny', 'DictAny',
)

import sys

from os import PathLike
from typing import TYPE_CHECKING
from typing import List, Dict, AnyStr, Any
from typing import Optional, Union, Tuple, Mapping, Sequence, TypeVar, Type
from typing import Callable, Coroutine, Generator, IO, Iterable, Iterator, AsyncIterator
from typing import cast, overload

if sys.version_info >= (3, 8):
    from typing import Final
else:
    from typing_extensions import Final

Data = TypeVar('Data', str, List[str], Dict[str, Union[str, List[str]]])
AnyMany = TypeVar('AnyMany', Any, List[Any])

TextMany = TypeVar('TextMany', str, List[str])
TextList = List[str]

DictList = List[Dict[str, Any]]
DictMany = TypeVar('DictMany', Dict[str, Any], List[Dict[str, Any]])
DictAny = Dict[str, Any]
DictText = Dict[str, str]



"""
Patch Functions to modify pathlib.Path

borrowed some functions from https://github.com/KenKundert/extended_pathlib/blob/master/extended_pathlib.py
"""

import os
import sys
import codecs
import aiofiles
import tempfile
from pathlib import Path, PosixPath
from lazycls.prop import classproperty
from typing import TypeVar, Union, Type, List, Any, Optional, Callable, Dict


PathLike = TypeVar('PathLike', type(Path), Path, PosixPath, os.PathLike, str, Union[str, Path], Union[str, PosixPath], Union[str, os.PathLike])


def _is_readable(path):
    """
    Tests whether path exists and is readable.
    >>> from extended_pathlib import Path
    >>> Path('/usr/bin/python').is_readable()
    True
    """
    return os.access(str(path), os.R_OK)


PosixPath.is_readable = _is_readable


# is_writable {{{1
def _is_writable(path):
    """
    Tests whether path exists and is writable.
    >>> Path('/usr/bin/python').is_writable()
    False
    """
    return os.access(str(path), os.W_OK)


PosixPath.is_writable = _is_writable


# is_executable {{{1
def _is_executable(path):
    """
    Tests whether path exists and is executable.
    >>> Path('/usr/bin/python').is_executable()
    True
    """
    return os.access(str(path), os.X_OK)


PosixPath.is_executable = _is_executable


# is_hidden {{{1
def _is_hidden(path):
    """
    Tests whether path exists and is hidden.
    >>> Path('/usr/bin/python').is_hidden()
    False
    """
    return path.exists() and path.name.startswith(".")


PosixPath.is_hidden = _is_hidden


# is_newer {{{1
def _is_newer(path, ref):
    """
    Tests whether path is newer than ref where ref is either another path or a
    date.
    >>> Path('/usr/bin/python').is_newer(0)
    True
    """
    mtime = path.stat().st_mtime
    try: return mtime > ref
    except TypeError:
        try: return mtime > ref.timestamp
        except AttributeError: return mtime > ref.stat().st_mtime


PosixPath.is_newer = _is_newer


# path_from {{{1
def _path_from(path, start):
    """
    Returns relative path from start as a path object.
    This differs from Path.relative_to() in that relative_to() will not return a
    path that starts with '..'.
    >>> Path('.').path_from('..')
    PosixPath('tests')
    """
    return Path(os.path.relpath(str(path), str(start)))


PosixPath.path_from = _path_from


# sans_ext {{{1
def _sans_ext(path):
    """
    Removes the file extension.
    This differs from Path.stem, which returns the final path component
    stripped of its extension. This returns the full path stripped of its
    extension.
    >>> Path('a/b.c').sans_ext()
    PosixPath('a/b')
    """
    return path.parent / path.stem


PosixPath.sans_ext = _sans_ext

"""
Pulled from lazycls.io
"""

def _get_cwd() -> str:
    return os.getcwd()

def _get_user_home_dir() -> str:
    return os.path.expanduser('~')

@classproperty
def _working_dir(cls) -> PathLike:
    return cls(_get_cwd())

Path.working_dir = _working_dir

@classproperty
def _userhome(cls) -> PathLike:
    """Return a new path pointing to the user's home directory (as
    returned by os.path.expanduser('~')).
    """
    return cls(cls()._flavour.gethomedir(None))

Path.userhome = _userhome


@property
def _string(self) -> str:
    """Returns string representation of Path.as_posix()"""
    return self.as_posix()

Path.string = _string

@property
def _extension(self):
    """ alias for Path.suffix or the file extension"""
    return self.suffix

Path.extension = _extension

@property
def _exist(self) -> bool:
    """property value of path.exists()"""
    return self.exists()

Path.exist = _exist

def _to_pathlike(path: PathLike) -> PathLike:
    if isinstance(path, str): path = Path(path)
    return path

Path.to_pathlike = _to_pathlike

def _to_str_from_path(path: PathLike) -> PathLike:
    if not isinstance(path, str): 
        try: return path.string
        except: return path.as_posix()
    return path

Path.to_pathstr = _to_str_from_path

def _resolve_rel_str_path(path: PathLike) -> str:
    pathstr = _to_str_from_path(path)
    if pathstr.startswith('~'): pathstr = pathstr.replace('~', _get_user_home_dir(), 1)
    #elif pathstr.startswith('../'): pathstr = _get_cwd() + pathstr
    elif pathstr.startswith('../'): pathstr = pathstr.replace('..', _get_cwd(), 1)
    #elif pathstr.startswith('..'): pathstr = _get_cwd() + '/' + pathstr
    #elif pathstr.startswith('..'): pathstr = pathstr.replace('..', Path(_get_cwd()).parent.parent.as_posix() + '/', 1)
    elif pathstr.startswith('..'): pathstr = pathstr.replace('..', Path(_get_cwd()).parent.parent.as_posix() + '/', 1)
    elif pathstr.startswith('./'): pathstr = pathstr.replace('.', _get_cwd(), 1)
    #elif pathstr.startswith('.'): pathstr = pathstr.replace('.', _get_cwd() + '/', 1)
    elif pathstr.startswith('.'): pathstr = pathstr.replace('.', Path(_get_cwd()).parent.as_posix() + '/', 1)
    return pathstr

Path.resolve_to_pathstr = _resolve_rel_str_path

def _resolve_rel_path(path: PathLike) -> PathLike:
    pathstr = _resolve_rel_str_path(path)
    return Path(pathstr)


Path.resolve_to_path = _resolve_rel_path

def _get_path(path: PathLike, resolve: bool = True, ensure_exists: bool = False, mode=0o777, file_mode=0o666, parents=False, exist_ok: bool = True) -> PathLike:
    path = _resolve_rel_path(path)
    #if isinstance(path, str): path = Path(path)
    if resolve: path.resolve()
    if ensure_exists and path.is_dir(): path.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
    elif ensure_exists: path.touch(mode=file_mode, exist_ok=exist_ok)
    return path

Path.get_path = _get_path

def _get_user_path(path: PathLike, resolve: bool = False) -> PathLike:
    if isinstance(path, str): path = Path(path)
    path = path.expanduser()
    if resolve: path.resolve()
    return path

Path.get_user_path = _get_user_path

def _from_env(key: str, default: PathLike, **kwargs) -> PathLike:
    """ gets the path from env value, returns default if none"""
    val = os.getenv(key, None)
    if val is None: return _get_path(default, **kwargs)
    return _get_path(val, **kwargs)

Path.from_env = _from_env

@staticmethod
def _get_parent_path(path: PathLike, *args, **kwargs) -> PathLike:
    p = _get_path(path, *args, **kwargs)
    return p.parent

Path.get_parent_path = _get_parent_path

@staticmethod
def _get_lib_path(posix: bool = False) -> PathLike:
    p = _get_cwd()
    if posix: return p
    return Path(p)

Path.get_lib_path = _get_lib_path

@staticmethod
def _get_cwd_paths(*paths, posix: bool = True) -> PathLike:
    p = _get_cwd()
    if not paths:
        if posix: return p
        return Path(p)
    if posix: return Path(p).joinpath(*paths).as_posix()
    return Path(p).joinpath(*paths)

Path.get_cwd_paths = _get_cwd_paths

"""
Open extensions
"""

def _reader(self, mode: str = 'r', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
    """ Alias for open(r)"""
    return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)

Path.reader = _reader
Path.read_mode = _reader
Path.in_reader = _reader

def _readerb(self, mode: str = 'rb', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
    """ Alias for open(rb)"""
    return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)

Path.readerb = _readerb
Path.reader_b = _readerb
Path.read_mode_b = _readerb
Path.binary_reader = _readerb
Path.b_reader = _readerb
Path.reader_binary = _readerb
Path.read_mode_binary = _readerb
Path.read_binary_mode = _readerb
Path.in_reader_b = _readerb

def _appender(self, mode: str = 'a', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
    """ Alias for open(a)"""
    return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)

Path.appender = _appender
Path.append_mode = _appender
Path.in_appender = _appender

def _appenderb(self, mode: str = 'ab', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
    """ Alias for open(ab)"""
    return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)

Path.appenderb = _appenderb
Path.appender_b = _appenderb
Path.append_mode_b = _appenderb
Path.binary_appender = _appenderb
Path.b_appender = _appenderb
Path.appender_binary = _appenderb
Path.append_mode_binary = _appenderb
Path.append_binary_mode = _appenderb
Path.in_appender_b = _appenderb

def _writer(self, mode: str = 'w', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
    """ Alias for open(w)"""
    return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)

Path.writer = _writer
Path.write_mode = _writer
Path.in_writer = _writer

def _writerb(self, mode: str = 'wb', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
    """ Alias for open(wb)"""
    return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)

Path.writerb = _writerb
Path.writer_b = _writerb
Path.write_mode_b = _writerb
Path.binary_writer = _writerb
Path.b_writer = _writerb
Path.writer_binary = _writerb
Path.write_mode_binary = _writerb
Path.write_binary_mode = _writerb
Path.in_writer_b = _writerb

from lazycls.serializers import Base, OrJson, Yaml, Pkl

_LOADERS = {
    '.json': OrJson,
    '.yml': Yaml,
    '.yaml': Yaml,
    '.pkl': Pkl,
    '.pickle': Pkl,
    '.pb': Pkl,
    'json': OrJson,
    'yaml': Yaml,
    'yml': Yaml,
    'pkl': Pkl,
    'pickle': Pkl
}

def _to_base64(self, **kwargs):
    """ Reads the file as text, then turns to base64"""
    data = self.read_text(**kwargs)
    return Base.b64_encode(data)

Path.to_base64 = _to_base64
Path.to_b64 = _to_base64

def _to_gzip_base64(self, **kwargs):
    """ Reads the file as text, then turns to gzip+base64"""
    data = self.read_text(**kwargs)
    return Base.b64_gzip_encode(data)
    
Path.to_gzip_base64 = _to_gzip_base64
Path.to_bgz = _to_gzip_base64

@staticmethod
def _from_base64(data: Union[str, bytes], path: PathLike = None, encoding: str = 'utf-8', errors: str = None, **kwargs) -> PathLike:
    d = Base.b64_decode(data)
    if not path: return d
    p = _get_path(path, **kwargs)
    p.write_text(data=d, encoding=encoding, errors=errors)
    return p

Path.from_base64 = _from_base64
Path.from_b64 = _from_base64

@staticmethod
def _from_gzip_base64(data: Union[str, bytes], path: PathLike = None, encoding: str = 'utf-8', errors: str = None, **kwargs) -> PathLike:
    d = Base.b64_gzip_decode(data)
    if not path: return d
    p = _get_path(path, **kwargs)
    p.write_text(data=d, encoding=encoding, errors=errors)
    return p

Path.from_gzip_base64 = _from_gzip_base64
Path.from_bgz = _from_gzip_base64

def _read_bytes(self):
    """
    Open the file in bytes mode (rb), read it, and close the file.
    """
    with self.open(mode='rb') as f:
        return f.read()

Path.read_bytes = _read_bytes

def _read_text(self, encoding: str = 'utf-8', errors: str = None):
    """
    Open the file in text mode (r), read it, and close the file.
    """
    with self.open(mode='r', encoding=encoding, errors=errors) as f:
        return f.read()

Path.read_text = _read_text

def _read_text_to_lines(self, encoding: str = 'utf-8', errors: str = None, newlines: str = '\n', striplines: bool = True):
    data = self.read_text(encoding=encoding, errors=errors)
    lines = data.split(newlines)
    if striplines: lines = [l.strip() for l in lines]
    return lines

Path.read_textlines = _read_text_to_lines

def _read_json(self, encoding: str = 'utf-8', errors: str = None, **kwargs):
    data = self.read_text(encoding=encoding, errors=errors)
    return OrJson.loads(data, **kwargs)
    
Path.read_json = _read_json

def _read_yaml(self, encoding: str = 'utf-8', errors: str = None, **kwargs):
    data = self.read_text(encoding=encoding, errors=errors)
    return Yaml.loads(data, **kwargs)

Path.read_yaml = _read_yaml

def _read_pickle(self, encoding: str = 'utf-8', errors: str = None, **kwargs):
    data = self.read_bytes(encoding=encoding, errors=errors)
    return Pkl.loads(data, **kwargs)

Path.read_pickle = _read_pickle
Path.read_pkl = _read_pickle


def _append_bytes(self, data):
    """
    Open the file in bytes mode (ab), append to it, and close the file.
    """
    # type-check for the buffer interface before truncating the file
    view = memoryview(data)
    with self.open(mode='ab') as f:
        return f.write(view)

Path.append_bytes = _append_bytes

def _append_text(self, data: str, encoding: str ='utf-8', errors: str = None, insert_newline: bool = True):
    """
    Open the file in text mode (a), write to it, and close the file.
    """
    if not isinstance(data, str): raise TypeError('data must be str, not %s' % data.__class__.__name__)
    with self.open(mode='a', encoding=encoding, errors=errors) as f:
        f.write(data)
        if insert_newline: f.write('\n')
        return f

Path.append_text = _append_text


def _write_text(self, data: str, encoding: str ='utf-8', errors: str = None):
    """
    Open the file in text mode (w), write to it, and close the file.
    """
    if not isinstance(data, str): raise TypeError('data must be str, not %s' % data.__class__.__name__)
    with self.open(mode='w', encoding=encoding, errors=errors) as f:
        return f.write(data)

Path.write_text = _write_text

def _write_bytes(self, data):
    """
    Open the file in bytes mode (wb), write to it, and close the file.
    """
    # type-check for the buffer interface before truncating the file
    view = memoryview(data)
    with self.open(mode='wb') as f:
        return f.write(view)

Path.write_bytes = _write_bytes

def _write_json(self, data: Any, encoding: str = 'utf-8', errors: str = None, **kwargs):
    return self.write_text(OrJson.dumps(data, **kwargs), encoding=encoding, errors=errors)

Path.write_json = _write_json

def _write_yaml(self, data: Any, encoding: str = 'utf-8', errors: str = None, **kwargs):
    return self.write_text(Yaml.dumps(data, **kwargs), encoding=encoding, errors=errors)

Path.write_yaml = _write_yaml

def _write_pickle(self, data: Any, **kwargs):
    return self.write_bytes(Pkl.dumps(data, **kwargs))

Path.write_pickle = _write_pickle
Path.write_pkl = _write_pickle

def _get_loader(self, loader: str = None):
    """ If the file is supported (pkl, yaml, json) - then return the loader"""
    if loader: return _LOADERS.get(loader, _LOADERS.get(self.suffix, None))
    return _LOADERS.get(self.extension, None)

Path.get_loader = _get_loader

def _loads(self, encoding: str ='utf-8', errors: str = None, loader: str = None, binary: bool = False, **kwargs) -> Any:
    """ If the file is supported (pkl, yaml, json) - then return the loader.loads(**kwargs) method.
        else:
            - if binary: returns .read_bytes()
                else: returns .read_text()
    """
    l = self.get_loader(loader=loader)
    if l: return l.loads(self.read_text(encoding=encoding, errors=errors), **kwargs)
    if binary: return self.read_bytes()
    return self.read_text(encoding=encoding, errors=errors)

Path.loads = _loads

def _dumps(self, data: Any, encoding: str ='utf-8', errors: str = None, loader: str = None, **kwargs) -> Any:
    """ If the file is supported (pkl, yaml, json) - then perform the loader.dumps(data, **kwargs) method.
        else:
            - if binary: performs .write_bytes(data)
                else: performs .write_text(data)
        returns: Path
    """
    l = self.get_loader(loader=loader)
    if l:
        d = l.dumps(data, **kwargs)
        if isinstance(d, bytes): self.write_bytes(d)
        else: self.write_text(d, encoding=encoding, errors=errors)
    elif isinstance(data, bytes): self.write_bytes(data)
    else: self.write_text(data, encoding=encoding, errors=errors)
    return self

Path.dumps = _dumps

def _get_files(self, pattern: str = '*') -> List[PathLike]:
    """
    Uses the .glob(pattern) to construct List[PathLike] instead of returning an iterator
    """
    p_iter = self.glob(pattern)
    return [p for p in p_iter if p.is_file]

Path.get_files = _get_files

@staticmethod
def _get_tempfile(mode: str = 'w', encoding: str = 'utf-8', dir: Any = None, buffering: int = -1, newline: str = None, delete: bool = False, **kwargs):
    return tempfile.NamedTemporaryFile(mode=mode, encoding=encoding, buffering=buffering, newline=newline, delete=delete, dir=dir, **kwargs)

Path.get_tempfile = _get_tempfile

@staticmethod
def _get_tempdir(dir: Any = None, **kwargs):
    return tempfile.TemporaryDirectory(dir=dir, **kwargs)

Path.get_tempdir = _get_tempdir

"""
Async Functions and Extensions
"""

from lazycls.ext._imports import LazyLib

@classproperty
def _async_os(cls):
    """ Aliases for aiofiles.os"""
    return aiofiles.os

Path.async_os = _async_os

async def _async_open(self, mode: str = 'r', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
    """ Alias for aiofiles.open(r)"""
    return await aiofiles.open(self.as_posix(), mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)

Path.async_open = _async_open


async def _async_read(self, mode: str = 'r', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
    """ Alias for aiofiles.open(r)"""
    return await aiofiles.open(self.as_posix(), mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)

Path.async_reader = _async_read
Path.async_read_mode = _async_read

async def _async_read_b(self, mode: str = 'rb', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
    """ Alias for aiofiles.open(rb)"""
    return await aiofiles.open(self.as_posix(), mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)

Path.async_reader_b = _async_read_b
Path.async_read_mode_b = _async_read_b

async def _async_append(self, mode: str = 'a', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
    """ Alias for aiofiles.open(a)"""
    return await aiofiles.open(self.as_posix(), mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)

Path.async_appender = _async_append
Path.async_append_mode = _async_append

async def _async_append_b(self, mode: str = 'ab', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
    """ Alias for aiofiles.open(ab)"""
    return await aiofiles.open(self.as_posix(), mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)

Path.async_appender_b = _async_append_b
Path.async_append_mode_b = _async_append_b

async def _async_write(self, mode: str = 'w', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
    """ Alias for aiofiles.open(w)"""
    return await aiofiles.open(self.as_posix(), mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)

Path.async_writer = _async_write
Path.async_write_mode = _async_write

async def _async_write_b(self, mode: str = 'wb', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
    """ Alias for aiofiles.open(wb)"""
    return await aiofiles.open(self.as_posix(), mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)

Path.async_writer_b = _async_write_b
Path.async_write_mode_b = _async_write_b

async def _async_read_bytes(self, **kwargs):
    """
    Open the file in bytes mode (rb), read it, and close the file.
    """
    async with self.async_open(mode='rb', **kwargs) as f:
        return await f.read()

Path.async_read_bytes = _async_read_bytes

async def _async_read_text(self, encoding: str = 'utf-8', errors: str = None, **kwargs):
    """
    Open the file in text mode (r), read it, and close the file.
    """
    async with self.async_open(mode='r', encoding=encoding, errors=errors, **kwargs) as f:
        return await f.read()

Path.async_read_text = _async_read_text

@staticmethod
async def _async_get_tempfile(mode: str = 'w', encoding: str = 'utf-8', dir: Any = None, buffering: int = -1, newline: str = None, delete: bool = False, **kwargs):
    return await aiofiles.tempfile.NamedTemporaryFile(mode=mode, encoding=encoding, buffering=buffering, newline=newline, delete=delete, dir=dir, **kwargs)

Path.async_get_tempfile = _async_get_tempfile

@staticmethod
async def _async_get_tempdir(dir: Any = None, **kwargs):
    return await aiofiles.tempfile.TemporaryDirectory(dir=dir, **kwargs)

Path.async_get_tempdir = _async_get_tempdir

if LazyLib.is_avail_requests:
    _reqs = LazyLib.requests
    
    @staticmethod
    def _from_url(url: str, path: PathLike, overwrite: bool = False, **kwargs) -> PathLike:
        """ Downloads a file from url and saves to  file """
        p = _get_path(path, resolve=True)
        if p.exists() and not overwrite: return p
        r = _reqs.get(url=url, **kwargs)
        if r.status_code >= 400: raise Exception(f'url = {url} returned status_code = {r.status_code}')
        p.write_bytes(data=r.content)
        return p
    
    Path.from_url = _from_url   

if LazyLib.is_avail_httpx:
    _httpx = LazyLib.httpx
    
    @staticmethod
    async def _async_from_url(url: str, path: PathLike, overwrite: bool = False, client_args: Dict[str, Any] = {}, **kwargs) -> PathLike:
        """ Downloads a file from url and saves to  file """
        p = _get_path(path, resolve=True)
        if p.exists() and not overwrite: return p
        async with _httpx.AsyncClient(url, **client_args) as c:
            r = await c.get(url=url, **kwargs)
        if r.status_code >= 400: raise Exception(f'url = {url} returned status_code = {r.status_code}')
        p.write_bytes(data=r.content)
        return p

    Path.async_from_url = _async_from_url




import os
from collections import namedtuple
from inspect import currentframe
from pydantic import BaseSettings
from pydantic.env_settings import SettingsSourceCallable


try:
    from pydantic.main import ModelMetaclass
except ImportError:  # pragma: no cover
    ModelMetaclass = type(BaseSettings)

from lazycls.typec import *
from lazycls.ext.pathio import Path, PathLike
from lazycls.serializers import OrJson

class EnvClsMeta(ModelMetaclass):
    def __new__(mcs, name, bases, namespace):
        Config = namespace.setdefault("Config", type("Config", (), {}))
        if not hasattr(Config, "env_prefix"):
            #if "base" in name.lower() or "config" in name.lower(): Config.env_prefix = ""
            #if "base" in name.lower() or "settings" in name.lower(): Config.env_prefix = ""
            if "base" in name.lower(): Config.env_prefix = ""
            elif "Cfg" in name: Config.env_prefix = name.replace("Cfg", "") + "_"
            elif "Config" in name: Config.env_prefix = name.replace("Config", "") + "_"
            elif "Settings" in name: Config.env_prefix = name.replace("Settings", "") + "_"
            else: Config.env_prefix = name + "_"
            
        cls = super().__new__(mcs, name, bases, namespace)
        if not getattr(Config, "auto_init", True): return cls
        else: return cls()


def dynamic_config_settings_source(settings: BaseSettings) -> Dict[str, Any]:
    """
    Settings source that loads variables from a .json/.yaml/.env/.txt file
    using cls.config.env_prefix.upper() + 'CONFIG_FILE' environment variable
    to determine the env file location.
    
    i.e. class CustomConfig(EnvCls) = 'CUSTOM_CONFIG_FILE' -> /path/to/.env
    """
    encoding = settings.__config__.env_file_encoding
    config_file_env_name = settings.__config__.env_prefix.upper() + 'CONFIG_FILE'
    config_file_env_name = config_file_env_name.replace('CONFIG_CONFIG',' CONFIG_', 1) # Prevents duplicate config names 
    config_file_val = os.getenv(config_file_env_name)
    if not config_file_val: return {}
    p = Path(config_file_val)
    if not p.exists(): return {}
    if p.extension in {'.yml', '.yaml'}: return p.read_yaml(encoding=encoding)
    if p.extension == '.json': return p.read_json(encoding=encoding)
    if p.extension in {'.pickle', '.pkl'}: return p.read_pkl()
    from dotenv import dotenv_values
    return dotenv_values(p.string, encoding=encoding)
    
class EnvCls(BaseSettings, metaclass=EnvClsMeta):

    class Config:
        @classmethod
        def customise_sources(cls, init_settings: SettingsSourceCallable, env_settings: SettingsSourceCallable, file_secret_settings: SettingsSourceCallable) -> Tuple[SettingsSourceCallable, ...]:
            return init_settings, dynamic_config_settings_source, env_settings, file_secret_settings

        auto_init = False
        env_prefix = ""
        arbitrary_types_allowed = True
        extra = 'allow'
        json_loads = OrJson.loads
        json_dumps = OrJson.dumps


    @property
    def env_values(self): 
        return {
            self.Config.env_prefix.upper() + k.upper(): v
            for k, v in self.dict().items()
        }

    def prefixed_dict(self, *args, **kwargs) -> Dict[str, Any]:
        """
        Similar to pydantic's `dict()`, but the `env_prefix` is included in the keys,
        so they match the original environment variable names.
        For example:
            class Database(EnvCls):
                host = "localhost"
                username = "admin"
            assert Database.dict() == {"host": "localhost", "username": "admin"}
            assert Database.prefixed_dict() == {"DATABASE_HOST": "localhost", "DATABASE_USERNAME": "admin"}
        """
        return {
            self.Config.env_prefix.upper() + k.upper(): v
            for k, v in self.dict(*args, **kwargs).items()
        }

    def __call__(self, *args, **kwargs) -> "EnvCls":
        """
        Instantiate this class as if you were calling the class itself.
        """
        return type(self)(*args, **kwargs)
    
    def reload(self, *args, **kwargs):
        """
        Reloads the class by reinitializing and reloading environment variables
        if they've changed.
        
        os.environ['DATABASE_HOST'] = 'localhost'
        assert Database.host == 'localhost'
        os.environ['DATABASE_HOST'] = '127.0.0.1'
        Database.reload()
        assert Database.host == '127.0.0.1'
        """
        self = self.__init__(*args, **kwargs)


def populate_globals(globs=None):
    """
    Search for instances of `EnvCls` in the global variables in the calling context
    and then update the global variables with the `prefixed_dict()` of those `EnvCls` isntances.
    For example, this code::
        class Database(EnvCls):
            host = "localhost"
            username = "admin"
        populate_globals()
    will result in global variables `DATABASE_HOST` and `DATABASE_USERNAME`.
    This is useful in e.g. Django where settings need to be declared at the global level.
    You can pass your own dict for the function to use instead of the current global variables.
    """
    if globs is None:
        globs = currentframe().f_back.f_globals

    for cls in list(globs.values()):
        if isinstance(cls, EnvCls):
            globs.update(cls.prefixed_dict())



class CmdError(Exception):
    # https://www.kbairak.net/programming/python/2021/01/21/custom_exceptions.html
    def __init__(self, returncode, stdout, stderr):
        super().__init__(returncode, stdout, stderr)

    @property
    def returncode(self):
        return self.args[0]

    @property
    def stdout(self):
        return self.args[1]

    @property
    def stderr(self):
        return self.args[2]



import os
import typing
import pathlib
from typing import Callable, Dict, Tuple, Type, Union, TypeVar, List

from . import core
from . import cloud

PathLike = core.PathLike
ReadOnlyPath = core.ReadOnlyPath
ReadWritePath = core.ReadWritePath
PathLikeCls = Union[Type[ReadOnlyPath], Type[ReadWritePath]]

T = TypeVar('T')

_PATHLIKE_CLS: Tuple[PathLikeCls, ...] = (
    core.PosixFSxPath,
    core.WindowsFSxPath,
    cloud.PosixGCSPath,
    cloud.PosixS3Path,
    cloud.PosixMinioPath,
    cloud.PosixS3CompatPath,
    cloud.WindowsGCSPath,
    cloud.WindowsS3Path,
    cloud.WindowsMinioPath,
    cloud.WindowsS3CompatPath,
)

_URI_PREFIXES_TO_CLS: Dict[str, PathLikeCls] = {
    'gs://': cloud.PosixGCSPath,
    's3://': cloud.PosixS3Path,
    'minio://': cloud.PosixMinioPath,
    's3compat://': cloud.PosixS3CompatPath,
}


@typing.overload
def register_pathlike_cls(path_cls_or_uri_prefix: str) -> Callable[[T], T]:
    ...


@typing.overload
def register_pathlike_cls(path_cls_or_uri_prefix: T) -> T:
    ...


def register_pathlike_cls(path_cls_or_uri_prefix):
    global _PATHLIKE_CLS
    if isinstance(path_cls_or_uri_prefix, str):

        def register_pathlike_decorator(cls: T) -> T:
            _URI_PREFIXES_TO_CLS[path_cls_or_uri_prefix] = cls
            return register_pathlike_cls(cls)

        return register_pathlike_decorator
    else:
        _PATHLIKE_CLS = _PATHLIKE_CLS + (path_cls_or_uri_prefix,)
        return path_cls_or_uri_prefix


def as_path(path: PathLike) -> ReadWritePath:
    """Create a generic `pathlib.Path`-like abstraction.
    Depending on the input (e.g. `gs://`, `github://`, `ResourcePath`,...), the
    system (Windows, Linux,...), the function will create the right pathlib-like
    abstraction.
    Args:
        path: Pathlike object.
    Returns:
        path: The `pathlib.Path`-like abstraction.
    """
    is_windows = os.name == 'nt'
    if isinstance(path, str):
        uri_splits = path.split('://', maxsplit=1)
        if len(uri_splits) > 1:    # str is URI (e.g. `gs://`, `github://`,...)
            # On windows, `PosixGCSPath` is created for `gs://` paths
            return _URI_PREFIXES_TO_CLS[uri_splits[0] + '://'](path)
        elif is_windows: return cloud.WindowsGCSPath(path)
        else: return core.PosixFSxPath(path)
    elif isinstance(path, _PATHLIKE_CLS):
        return path
    elif isinstance(path, os.PathLike): # Other `os.fspath` compatible objects
        path_cls = core.WindowsFSxPath if is_windows else core.PosixFSxPath
        return path_cls(path)
    else: raise TypeError(f'Invalid path type: {path!r}')


def get_userhome(as_pathz: bool = True):
    h = os.path.expanduser('~')
    if as_pathz: return as_path(h)
    return h

def get_cwd():
    return os.getcwd()

def resolve_relative(filepath: Union[str, PathLike]) -> str:
    if not isinstance(filepath, str): filepath = filepath.as_posix()
    if '://' in filepath: return filepath
    if filepath.startswith('~'): filepath = filepath.replace('~', get_userhome(), 1)
    elif filepath.startswith('../'): filepath = filepath.replace('..', get_cwd(), 1)
    elif filepath.startswith('..'): filepath = filepath.replace('..', pathlib.Path(get_cwd()).parent.parent.as_posix() + '/', 1)
    elif filepath.startswith('./'): filepath = filepath.replace('.', get_cwd(), 1)
    elif filepath.startswith('.'): filepath = filepath.replace('.', pathlib.Path(get_cwd()).parent.as_posix() + '/', 1)
    return filepath

def get_path(filepath: Union[str, PathLike], resolve: bool = False) -> _PATHLIKE_CLS:
    if resolve: filepath = resolve_relative(filepath)
    if isinstance(filepath, str): filepath = as_path(filepath)
    return filepath

def get_pathlike(filepath: Union[str, PathLike], resolve: bool = False) -> _PATHLIKE_CLS:
    if resolve: filepath = resolve_relative(filepath)
    if isinstance(filepath, str): filepath = as_path(filepath)
    return filepath

def get_lazydir(mkdir: bool = False):
    """
    returns resolved '~/.lazy'
    """
    p = get_userhome()
    p = p.joinpath('.lazy')
    if mkdir: p.mkdir(exist_ok=True, parents=True)
    return p



from .core import *
from ._json import JsonBase, OrJson, SimdJson, Json
from ._yaml import Yaml
from ._base import Base
from ._secrets import Secret
from ._pickle import BasePickle, Pickle, Dill, Pkl, Compression
from ._multi import YamlBase64, YamlBGZ, JsonBase64, JsonBGZ

class Serialize:
    Json: JsonBase = Json
    JsonB64: JsonBase64 = JsonBase64
    JsonBGZ: JsonBGZ = JsonBGZ
    Yaml: Yaml = Yaml
    YamlB64: YamlBase64 = YamlBase64
    YamlBGZ: YamlBGZ = YamlBGZ
    Base: Base = Base
    Secret: Secret = Secret
    Pkl: BasePickle = Pkl
    Compress = Compression
    


    
__all__ = (
    'OrJson', 'SimdJson', 'Json',
    'Yaml',
    'Base',
    'Secret',
    'Pickle', 'Dill', 'Pkl',
    'Compression',
    'Serialize',
    'YamlBase64', 'YamlBGZ', 'JsonBase64', 'JsonBGZ'
)




""" Manages the Fuzes using Cachez """
import os
import signal
import pathlib
import threading
import multiprocessing


from types import ModuleType
from typing import Dict, Type, Any, Optional, Union

from lazy.models import BaseCls
from lazy.utils import _get_logger

try: from lazy.io import cachez
except ImportError: cachez = ModuleType
from lazy.types import classproperty
from lazy.configz.common import FuzeConfigz

from .generic import get_fuze, FuzeTypes
from . import utils

logger = _get_logger('Fuze')



class FuzeMount(object):
    def __init__(self, source: str, mount_target: str):
        self._source = source
        self._mount_target = mount_target
        self._mount_path = pathlib.Path(mount_target)
        #self._do_cleanup = cleanup
        self._fs = None
        self._fuzer_name = None
        self._process: Union[threading.Thread, multiprocessing.Process] = None
        self._alive = False
    
    @property
    def source(self): return self._source
    @property
    def mount_target(self): return self._mount_target

    @property
    def mount_path(self): return self._mount_path.as_posix()

    @property
    def source_path(self):
        uri_splits = self._source.split('://', maxsplit=1)
        if len(uri_splits) > 1: return uri_splits[-1]        
        return self._source

    def _cleanup(self):
        if self._mount_path.exists():
            utils.umount(self.mount_path)
            try: self._mount_path.unlink()
            except: pass

    def _prestart(self):
        self._cleanup()
        self._mount_path.mkdir(exist_ok=True, parents=True)
    
    def _run_background(self):
        from .ops import FUZE

        func = lambda: FUZE(
            _FUSEr(fs, self.source_path, ready_file=ready_file), 
            self.target_path_str, 
            nothreads=not threads, 
            foreground=foreground)

    def _start_process(self, daemon: bool = True, **kwargs):
        assert not self._process, 'Process Active. Unable to Start'
        mp = utils.get_multiproc()
        self._process = mp.Process(target=self._run_background, daemon = daemon,  **kwargs)
        self._alive = True

    def _start_thread(self, daemon: bool = True, **kwargs):
        assert not self._process, 'Process Active. Unable to Start'

        mp = utils.get_multiproc()
        self._process = mp.Process(target=self._run_background, daemon = daemon,  **kwargs)
        self._alive = True

    def start(self, use_mp: bool = True, daemon: bool = True, **kwargs):
        if use_mp: return self._start_process(daemon = daemon, **kwargs)
        


    @property
    def fs(self) -> FuzeTypes:
        if not self._fs: 
            self._fs = get_fuze(self._source)
            self._fuzer_name = self._fs.__class__.__name__
            logger.info(f'Initialized Fuzer {self._fuzer_name} for {self._source}')
        return self._fs
    
    @property
    def process_id(self) -> Optional[int]:
        if not self._process: return None
        return self._process.pid


    @property
    def data(self) -> Dict[str, Any]:
        return {
            'fuzer': self._fuzer_name,
            'source': self._source,
            'source_path': self.source_path,
            'mount_target': self._mount_target,
            'mount_path': self.mount_path,
            'pid': self.process_id,
            'alive': self._alive,
        }


"""
Stores in a temp sqllite in ~/.cachez/cache.db

Stores a dict of:
mount_target: {
    'fuzer': str,
    'source': source,
    'mount_target': mount_target,
    'pid': process_id,
    'alive': bool,
},
source: {
    'fuzer': str,
    'source': source,
    'mount_target': mount_target,
    'pid': process_id,
    'alive': bool,
},

"""

class FuzeManager:
    _cache: cachez.Cache = None
    _cachedir: str = None
    _mounts: Dict[str, FuzeMount] = {} # [source, FuzeMount]

    @classproperty
    def cache(cls):
        if not cls._cache:
            from lazy.io import cachez
            cls._cachedir = FuzeConfigz.cache_dir.as_posix()
            logger.info(f"Initialized FuzeManager Cache at {cls._cachedir}")
            cls._cache = cachez.Cache(directory=cls._cachedir)
        return cls._cache
    
    @classmethod
    def _kill_pid(cls, pid: int, sig: int = signal.SIGTERM):
        return os.kill(pid, sig)

    @classmethod
    def _cleanup_process(cls, src: Dict[str, Any]):
        try: utils.umount(src.get('mount_path', src.get('mount_target')))
        except Exception as e:
            logger.error(e)
        if src.get('pid'): cls._kill_pid(src['pid'])

    @classmethod
    def cleanup_mount(cls, mount_target: str = None, source: str = None, src: Dict[str, Any] = None):
        assert source is not None and mount_target is not None and src is not None, 'Must provide source, mount_target or src dict'
        if not src:
            src = cls.cache.get(source or mount_target)
        cls._cleanup_process(src)
        logger.info(f'Completed Cleanup for {src}')
        if source: 
            cls.cache.pop(source, None)
            cls._mounts.pop(source, None)
        if mount_target: 
            cls.cache.pop(mount_target, None)
            cls._mounts.pop(mount_target, None)

    @classmethod
    def get_mount(cls, source: str, mount_target: str, fix_errors: bool = True):
        if cls._mounts.get(source): return cls._mounts[source]
        if cls._mounts.get(mount_target): return cls._mounts[mount_target]
        if cls.cache.get(source) or cls.cache.get(mount_target):
            src = cls.cache.get(source) or cls.cache.get(mount_target)
            logger.warning(f'Source Exists in Cache: {src}')
            assert src.get('alive', False) or fix_errors, 'Exiting as Fix Errors = False or Mount Point is still active'
            cls.cleanup_mount(mount_target, source, src)
        mountz = FuzeMount(source, mount_target)
        cls.cache[mountz.source] = mountz.data
        cls.cache[mountz.mount_target] = mountz.data
        cls._mounts[mountz.source] = mountz
        return mountz


        









    






class MountPoint(BaseCls):
    fuzer: str # Name of the class
    source: str # User provided Mount Source
    mount_point: str # User provided mount point
    cleanup: bool = True # Cleanup dir after unmount
    thread: Optional[threading.Thread] = None
    process: Optional[multiprocessing.Process] = None
    alive: bool = True

    @property
    def source_path(self):
        uri_splits = self.source.split('://', maxsplit=1)
        if len(uri_splits) > 1: return uri_splits[-1].strip()
        return self.source

    @property
    def target_path(self):
        return pathlib.Path(self.mount_point).resolve()
    
    @property
    def target_path_str(self):
        return self.target_path.as_posix()

    def kill(self, timeout: int = 5):
        if self.thread: self.thread.join(timeout=timeout)
        elif self.process:
            os.kill(self.process.pid, signal.SIGTERM) 
            self.process.join(timeout=timeout)

        self.alive = False
        self.thread = None
        logger.info(f'[{self.fuzer}] Completed Unmount: {self.source}')
    
    def _background(self, fs: FuseSystemType, foreground: bool = True, threads: bool = False, ready_file: bool = False):
        self.alive = True
        from fsspec.fuse import FUSE as _FUSE
        from fsspec.fuse import FUSEr as _FUSEr
        func = lambda: _FUSE(
            _FUSEr(fs, self.source_path, ready_file=ready_file), 
            self.target_path_str, 
            nothreads=not threads, 
            foreground=foreground)
        
        while self.alive:
            try:
                logger.info(f'Started Mount {self.source} -> {self.target_path_str} in Background')
                func()
            except KeyboardInterrupt:
                logger.error('Got KeyboardInterrupt. Unmounting')
                self.kill()
                #_kill_proc(self.mount_point)
                break
            except Exception as e:
                logger.error(f'Got Exception {e}. Unmounting')
                self.kill()
                #_kill_proc(self.mount_point)
                break

    
    def mount(self, fs: FuseSystemType, foreground: bool = True, threads: bool = False, ready_file: bool = False, use_mp: bool = False, daemon: bool = True, ops_class: 'FUSEr' = None):
        #from fsspec.fuse import FUSE as _FUSE
        #from fsspec.fuse import FUSEr as _FUSEr
        #if not ops_class: ops_class = _FUSEr
        self.target_path.mkdir(parents=True, exist_ok=True)
        if use_mp:
            #with _lock.acquire():
            #with threading.Lock() as _lockz:
            _lock.acquire()
            self.process = mproc.Process(target=self._background, args=(fs, foreground, threads, ready_file,), daemon=daemon)
            self.process.start()
            _lock.release()

            #with threading.RLock() as _lock:
            #    _lock.acquire()
        elif not foreground:
            #with _lock.acquire():
            #with threading._RLock() as _lock:
            _lock.acquire()
            self.thread = threading.Thread(target=self._background, args=(fs, foreground, threads, ready_file,), daemon = daemon)
            self.thread.start()
            _lock.release()
        
        else:
            self._background(fs, foreground = foreground, threads = threads, ready_file = ready_file)
        



        """
        func = lambda: _FUSE(
            ops_class(fs, self.source_path, ready_file=ready_file), 
            self.target_path_str, 
            nothreads=not threads, 
            foreground=foreground)
        if not foreground:
            th = threading.Thread(target=func)
            th.daemon = True
            th.start()
            logger.info(f'Started Mount {self.source} -> {self.target_path_str} in Background')
            self.thread = th
            _add_proc(self.mount_point, self)

        else:  # pragma: no cover
            try:
                logger.info(f'Started Mount {self.source} -> {self.target_path_str} in Foreground')
                func()
            except KeyboardInterrupt:
                _kill_proc(self.mount_point)
                #sys.exit()
        """


    def unmount(self, timeout: int = 5):
        #if not self.alive and self.thread is None: return
        logger.info(f'[{self.fuzer}] Unmounting: {self.source} -> {self.target_path.as_posix()}')
        self.kill(timeout)
        if self.cleanup:
            logger.info(f'[{self.fuzer}] Cleaning Up Mount Point: {self.target_path.as_posix()}')
            self.target_path.unlink()




from . import pathz

from .pathz import get_path, as_path, to_path, PathLike, register_pathlike_cls, PathzPath, get_lazydir

## fuze_v2 doesn't work as expected
## falling back to using others
#from . import fuze_v2

#from .fuze_v2 import GCSFuze, S3Fuze, MinioFuze, S3CompatFuze
#from .fuze_v2 import FuzeTypes, get_fuze, autofuze_mount, autofuze_unmount


import os
from .prop import classproperty
from .serializers import OrJson, Yaml
from .funcs import doesTextMatch
from .utils import Path, toPath
from .types import *


def hasEnv(name: str) -> bool: return os.getenv(name, None) is not None

def getEnv(): return {k:v for k,v in os.environ.items()}

def envToBool(name: str, default: str = 'false'): return bool(os.getenv(name, default).lower() in {'true', '1', 'yes'})

def envToType(value: str = None, valType = str):
    if value is None or value in {'none', 'None'}: return None
    return valType(value)

def envToList(name: str, sep: str = ',', default: List[str] = []):
    d = os.getenv(name, None)
    if d: return d.split(sep)
    return default

def envToDict(name: str, sep: str = '=', default: Dict[str, Any] = {}):
    d = envToList(name)
    if not d: return default
    rez = {}
    for i in d:
        for k,v in i.split(sep, 1):
            rez[k] = v
    return rez

def envToFloat(name: str, default: float = 0.0):
    d = os.getenv(name, None)
    if d is None: return default
    return envToType(d, float)

def envToInt(name: str, default: int = 0):
    d = os.getenv(name, None)
    if d is None: return default
    return envToType(d, int)

def envInVals(name: str, checkVals: List[str] = [], exact: bool = False, **kwargs) -> bool:
    d = os.getenv(name, None)
    if d is None: return False
    return doesTextMatch(text=d, items=checkVals, exact=exact, **kwargs)

def envToStr(name: str, default: str = ''):
    d = os.getenv(name, None)
    if d is None: return default
    return d


def toEnv(name: str, value: Any, override: bool = False):
    if not hasEnv(name) or override: os.environ[name] = str(value)

_LoadedEnvFiles = set()

def loadEnvFile(path: Union[str, Path], override: bool = False):
    global _LoadedEnvFiles
    path = toPath(path, resolve=True)
    if not path.exists(): return None
    if path.name in _LoadedEnvFiles: return True
    loader = None
    if path.suffix == '.json': loader = OrJson.loads
    elif path.suffix == '.yaml': loader = Yaml.loads
    if loader is None: return False
    data = loader(path.read_text())
    for k,v in data.items():
        toEnv(name=k, value=v, override=override)
    _LoadedEnvFiles.add(path.name)
    return True

def setEnvDict(data: Dict[str, Any], override: bool = False):
    for k,v in data.items():
        toEnv(name=k, value=v, override=override)

load_env_file = loadEnvFile
to_env = toEnv
set_env_from_dict = setEnvDict

"""
New Class to encapsulate all the above functions, to allow a singular import
of lazycls.envs.Env

will later deprecate the above functions
"""

class Env:
    list_delimiter: str = ','
    dict_delimiter: str = '='


    @classproperty
    def dict(cls):
        return {k:v for k,v in os.environ.items()}
    
    @classproperty
    def keys(cls):
        return list(os.environ.keys())
    
    @classproperty
    def values(cls):
        return list(os.environ.values())
    
    @classmethod
    def has(cls, key: str) -> bool:
        return os.getenv(key, None) is not None

    @classmethod
    def get(cls, key: str, default: Any = None):
        return os.environ.get(key, default)
    
    @staticmethod
    def cast(value: str = None, val_type = str):
        if value is None or value in {'none', 'None'}: return None
        return val_type(value)
    
    @classmethod
    def to_bool(cls, key: str, default: str = 'false'):
        return bool(os.getenv(key, default).lower() in {'true', '1', 'yes'})

    @classmethod
    def to_str(cls, key: str, default: Any = ''):
        val = os.getenv(key, None)
        if val is None: return default
        return val
    
    @classmethod
    def to_str_env(cls, src_key: str, to_key: str, default: Any = '', override: bool = False):
        """ used to handle credentials like AWS Keys, etc without directly exposing known env keys
            i.e. map src_key = AWS_KEYID -> to_key = AWS_ACCESS_KEY_ID
        """
        val = cls.to_str(src_key, default)
        if val: cls.set_env(key=to_key, value=val, override=override)
        return val
    
    @classmethod
    def to_list(cls, key: str, default: List[str] = []) -> List[str]:
        val = os.getenv(key, None)
        if val: return val.split(cls.list_delimiter)
        return default
    
    @classmethod
    def to_dict(cls, key: str, default: Dict[str, Any] = {}) -> Dict[str, Any]:
        val = cls.to_list(key)
        if not val: return default
        rez = {}
        for i in val:
            for k,v in i.split(cls.dict_delimiter, 1):
                rez[k.strip()] = v.strip()
        return rez

    @classmethod
    def to_float(cls, key: str, default: float = 0.0):
        val = os.getenv(key, None)
        if val is None: return default
        return cls.cast(val, float)
    
    @classmethod
    def to_int(cls, key: str, default: float = 0):
        val = os.getenv(key, None)
        if val is None: return default
        return cls.cast(val, int)
    
    @classmethod
    def to_b64(cls, key: str, default: Any = ''):
        """ Base64 encoded """
        val = os.getenv(key, None)
        if val is None: return default
        from lazycls.serializers import Base
        return Base.b64_decode(val)

    @classmethod
    def to_b64_env(cls, src_key: str, to_key: str, default: Any = '', override: bool = False):
        """ used to handle credentials like AWS Keys, etc without directly exposing known env keys
            i.e. map src_key = AWS_KEYID -> to_key = AWS_ACCESS_KEY_ID
        """
        val = cls.to_b64(src_key, default)
        if val: cls.set_env(key=to_key, value=val, override=override)
        return val

    @classmethod
    def to_bgz(cls, key: str, default: Any = ''):
        """ Base64 + Gzip encoded """
        val = os.getenv(key, None)
        if val is None: return default
        from lazycls.serializers import Base
        return Base.b64_gzip_decode(val)
    
    @classmethod
    def to_bgz_env(cls, src_key: str, to_key: str, default: Any = '', override: bool = False):
        """ used to handle credentials like AWS Keys, etc without directly exposing known env keys
            i.e. map src_key = AWS_KEYID -> to_key = AWS_ACCESS_KEY_ID
        """
        val = cls.to_bgz(src_key, default)
        if val: cls.set_env(key=to_key, value=val, override=override)
        return val
    
    @classmethod
    def to_json(cls, src_key: str, to_key: str, to_path: Union[str, Path], default: str = '{}', override: bool = False, as_posix: bool = False):
        """ for things like GOOGLE_APPLICATION_CREDENTIALS that is json based
            we read the env src_key, attempt to load it and then dump it in to_path
            i.e. map src_key = ADC_DATA -> creates file at to_path -> sets env[to_key] = to_path (to_key = GOOGLE_APPLICATION_CREDENTIALS)
        """
        path = toPath(to_path, resolve=True)
        if not path.exists() or override: 
            data = os.getenv(src_key, default)
            from lazycls.serializers import OrJson
            path.write_text(OrJson.dumps(OrJson.loads(data)), encoding='utf-8')            
        cls.set_env(key=to_key, value=path.as_posix(), override=override)
        if as_posix: return path.as_posix()
        return path
    
    @classmethod
    def to_json_b64(cls, src_key: str, to_key: str, to_path: Union[str, Path], default: str = '{}', override: bool = False, as_posix: bool = False):
        """ for things like GOOGLE_APPLICATION_CREDENTIALS that is json based with base64 encoding
            we read the env src_key, attempt to load it and then dump it in to_path
            i.e. map src_key = ADC_DATA -> creates file at to_path -> sets env[to_key] = to_path (to_key = GOOGLE_APPLICATION_CREDENTIALS)
        """
        path = toPath(to_path, resolve=True)
        if not path.exists() or override: 
            from lazycls.serializers import OrJson, Base
            data = os.getenv(src_key, None)
            if data is not None: data = Base.b64_decode(data)
            else: data = default
            path.write_text(OrJson.dumps(OrJson.loads(data)), encoding='utf-8')
        cls.set_env(key=to_key, value=path.as_posix(), override=override)
        if as_posix: return path.as_posix()
        return path
    
    @classmethod
    def to_yaml(cls, src_key: str, to_key: str, to_path: Union[str, Path], default: str = '{}', override: bool = False, as_posix: bool = False):
        """ similar to to_json, just with yaml"""
        path = toPath(to_path, resolve=True)
        if not path.exists() or override: 
            data = os.getenv(src_key, default)
            from lazycls.serializers import Yaml
            path.write_text(Yaml.dumps(Yaml.loads(data)), encoding='utf-8')            
        cls.set_env(key=to_key, value=path.as_posix(), override=override)
        if as_posix: return path.as_posix()
        return path

    
    @classmethod
    def in_vals(cls, key: str, values: List[str] = [], exact: bool = False, **kwargs) -> bool:
        val = os.getenv(key, None)
        if val is None: return False
        return doesTextMatch(text=val, items=values, exact=exact, **kwargs)
    
    @classmethod
    def set_env(cls, key: str, value: Any, override: bool = False):
        if not cls.has(key) or override: os.environ[key] = str(value)
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any], override: bool = False):
        for key, val in data.items():
            cls.set_env(key=key, value=val, override=override)
    
    @classmethod
    def load_file(cls, path: Union[str, Path], override: bool = False):
        global _LoadedEnvFiles
        path = toPath(path, resolve=True)
        if not path.exists(): return None
        if path.name in _LoadedEnvFiles: return True
        loader = None
        if path.suffix == '.json': loader = OrJson.loads
        elif path.suffix == '.yaml': loader = Yaml.loads
        if loader is None: return False
        data = loader(path.read_text())
        for k,v in data.items():
            cls.set_env(name=k, value=v, override=override)
        _LoadedEnvFiles.add(path.name)
        return True

    @classmethod
    def encode_to_string(cls, value: Any):
        """ Transforms the value input into one that can be reloaded from env"""
        if not value: return 'none'
        if isinstance(value, str): return value
        if isinstance(value, type(Path)): return value.read_text(encoding='utf8')
        if isinstance(value, list): return cls.list_delimiter.join(value)
        if isinstance(value, int) or isinstance(value, float): return str(value)
        if isinstance(value, bool): return str(value).lower()
        if isinstance(value, dict):
            kvs = [f"{k}{cls.dict_delimiter}{v}" for k,v in value.items()]
            return cls.list_delimiter.join(kvs)
        return str(value)

    @classmethod
    def save_config(cls, configmap: Dict[str, Any], path: Union[str, Path]):
        path = toPath(path, resolve=True)
        cfg = {k: cls.encode_to_string(v) for k,v in configmap.items()}
        # now we encode it further
        from lazycls.serializers import Base, Yaml, OrJson
        cfg = {k: Base.b64_gzip_encode(v) for k,v in cfg.items()}
        path.write_text(Yaml.dumps(cfg) if path.suffix in {'.yml', '.yaml'} else OrJson.dumps(cfg))
        return path
    
    @classmethod
    def load_config(cls, path: Union[str, Path], set_as_env: bool = True, override: bool = False):
        # Should only expect our serialized format.
        path = toPath(path, resolve=True)
        assert path.exists(), f'Invalid Path: {path.as_posix()} does not exist.'
        from lazycls.serializers import Base, Yaml, OrJson        
        cfg = Yaml.loads(path.read_text()) if path.suffix in {'.yml', '.yaml'} else OrJson.loads(path.read_text())
        cfg = {k: Base.b64_gzip_decode(v) for k,v in cfg.items()}
        if set_as_env: 
            for k,v in cfg.items():
                cls.set_env(name=k, value=v, override=override)
        return cfg
        


__all__ = [
    'hasEnv',
    'getEnv',
    'envToBool',
    'envToType',
    'envToList',
    'envToDict',
    'envToFloat',
    'envToInt',
    'envInVals',
    'envToStr',
    'toEnv',
    'setEnvDict',
    'set_env_from_dict',
    'loadEnvFile',
    'load_env_file',
    'to_env',
    'Env'
]



import orjson as _orjson

from os import PathLike
from pydantic import BaseModel
from typing import Dict, Any, List, Union, overload
from .core import Serializer, Mode, logger
from ._pysimd import _simdjson, SimdObject, SimdArray, create_simdobj


class JsonBase(Serializer):
    default_value: Dict[Any, Any] = None
    async_supported: bool = True
    cloud_supported: bool = True
    
    @classmethod
    def dumps(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        return cls._encode(obj, *args, default = default, **kwargs)

    @classmethod
    def loads(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        return cls._decode(data, *args, **kwargs)
    
    @classmethod
    def _encode(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        raise NotImplementedError
        
    @classmethod
    def _decode(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        raise NotImplementedError
    
    @classmethod
    async def async_dumps(cls, obj: Dict[Any, Any], *args, default: Any = None, **kwargs) -> str:
        if not cls.async_supported: raise Exception
        return await cls._async_encode(obj, *args, default = default, **kwargs)

    @classmethod
    async def async_loads(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if not cls.async_supported: raise Exception
        return await cls._async_decode(data, *args, **kwargs)
    
    @classmethod
    async def _async_encode(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        raise NotImplementedError
    
    @classmethod
    async def _async_decode(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        raise NotImplementedError

    @classmethod
    def readlines(cls, filelike, as_iterable: bool = False, ignore_errors: bool = True, *args, **kwargs):
        if as_iterable:
            def jsonlines_iterator():
                for line in filelike:
                    try: yield cls.loads(line, *args, **kwargs)
                    except (StopIteration, KeyboardInterrupt, GeneratorExit): break
                    except Exception as e:
                        logger.error(e)
                        if not ignore_errors: raise e
            return jsonlines_iterator
        rez = []
        for line in filelike:
            try: rez.append(cls.loads(line, *args, **kwargs))
            except Exception as e:
                logger.error(e)
                if not ignore_errors: raise e
        return rez
    
    @classmethod
    async def async_readlines(cls, filelike, as_iterable: bool = False, ignore_errors: bool = True, *args, **kwargs):
        if as_iterable:
            async def jsonlines_iterator():
                for line in filelike:
                    try: yield await cls.async_loads(line, *args, **kwargs)
                    except (StopIteration, KeyboardInterrupt, GeneratorExit): break
                    except Exception as e:
                        logger.error(e)
                        if not ignore_errors: raise e
            return jsonlines_iterator
        rez = []
        for line in filelike:
            try: rez.append(await cls.async_loads(line, *args, **kwargs))
            except Exception as e:
                logger.error(e)
                if not ignore_errors: raise e
        return rez
    
    @classmethod
    def loadlines(cls, path: PathLike, mode: str = Mode.read, encoding: str = 'utf-8', as_iterable: bool = False, ignore_errors: bool = True, *args, **kwargs):
        from lazy.io import get_path
        p = get_path(path)
        if as_iterable:
            def jsonlines_iterator():
                with p.open(mode=mode, encoding=encoding) as f:
                    for line in f:
                        try: yield cls.loads(line, *args, **kwargs)
                        except (StopIteration, KeyboardInterrupt, GeneratorExit): break
                        except Exception as e:
                            logger.error(e)
                            if not ignore_errors: raise e
            return jsonlines_iterator                
        rez = []
        with p.open(mode=mode, encoding=encoding) as f:
            for line in f:
                try: rez.append(cls.loads(line, *args, **kwargs))
                except Exception as e:
                    logger.error(e)
                    if not ignore_errors: raise e
        return rez
    
    @classmethod
    async def async_loadlines(cls, path: PathLike, mode: str = Mode.read, encoding: str = 'utf-8', as_iterable: bool = False, ignore_errors: bool = True, *args, **kwargs):
        from lazy.io import get_path
        p = get_path(path)
        if as_iterable:
            async def jsonlines_iterator():
                async with p.async_open(mode=mode, encoding=encoding) as f:
                    for line in f:
                        try: yield await cls.async_loads(line, *args, **kwargs)
                        except (StopIteration, KeyboardInterrupt, GeneratorExit): break
                        except Exception as e:
                            logger.error(e)
                            if not ignore_errors: raise e
            return jsonlines_iterator                
        rez = []
        async with p.async_open(mode=mode, encoding=encoding) as f:
            for line in f:
                try: rez.append(await cls.async_loads(line, *args, **kwargs))
                except Exception as e:
                    logger.error(e)
                    if not ignore_errors: raise e
        return rez

    @classmethod
    def iterlines(cls, path: PathLike, mode: str = Mode.read, encoding: str = 'utf-8', ignore_errors: bool = True, *args, **kwargs):
        from lazy.io import get_path
        p = get_path(path)
        with p.open(mode=mode, encoding=encoding) as f:
            for line in f:
                try: yield cls.loads(line, *args, **kwargs)
                except StopIteration: break
                except Exception as e:
                    logger.error(e)
                    if not ignore_errors: raise e
    
    @classmethod
    async def async_iterlines(cls, path: PathLike, mode: str = Mode.read, encoding: str = 'utf-8', ignore_errors: bool = True, *args, **kwargs):
        from lazy.io import get_path
        p = get_path(path)
        async with p.async_open(mode=mode, encoding=encoding) as f:
            for line in f:
                try: yield await cls.async_loads(line, *args, **kwargs)
                except StopIteration: break
                except Exception as e:
                    logger.error(e)
                    if not ignore_errors: raise e



class OrJson(JsonBase):
    
    @classmethod
    def _encode(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        return _orjson.dumps(obj, default=default, *args, **kwargs).decode()
    
    @classmethod
    def _decode(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        return _orjson.loads(data, *args, **kwargs)
    
    @classmethod
    async def _async_encode(cls, obj: Dict[Any, Any], *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        return _orjson.dumps(obj, default=default, *args, **kwargs).decode()
    
    @classmethod
    async def _async_decode(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        return _orjson.loads(data, *args, **kwargs)


class SimdJson(JsonBase):
    parser: _simdjson.Parser = _simdjson.Parser()
    parser_enabled: bool = True
    
    @classmethod
    def _encode(cls, obj: Any, *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        if isinstance(obj, (SimdObject, SimdArray)): obj = obj.data
        return _simdjson.dumps(obj, default=default, *args, **kwargs)
    
    @classmethod
    def _decode(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if cls.parser_enabled: return create_simdobj(cls.parser.parse(data, *args, **kwargs))
        return _simdjson.loads(data, *args, **kwargs)
    
    @classmethod
    def parse(cls, data: Any, *args, **kwargs) -> Union[Union[_simdjson.Object, _simdjson.Array], Union[Dict[Any, Any], List[str]]]:
        if not isinstance(data, (str, bytes)) and getattr(data, 'content', None): return create_simdobj(cls.parser.parse(data.content, *args, **kwargs))
        return create_simdobj(cls.parser.parse(data, *args, **kwargs))
    
    @classmethod
    def decode(cls, data: Any, *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if isinstance(data, (dict, list, set)): return data
        if issubclass(data, BaseModel): return data.dict()
        #if issubclass(data, _simdjson.Object) or isinstance(data, _simdjson.Object): return data.as_dict()
        #if issubclass(data, _simdjson.Array) or isinstance(data, _simdjson.Array): return data.as_list()
        if isinstance(data, _simdjson.Object): return data.as_dict()
        if isinstance(data, _simdjson.Array): return data.as_list()
        if isinstance(data, (SimdObject, SimdArray)): return data.data
        if isinstance(data, (str, bytes)): return _simdjson.loads(data)
        raise ValueError
    
    @classmethod
    async def _async_encode(cls, obj: Any, *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        if isinstance(obj, (SimdObject, SimdArray)): obj = obj.data
        return _simdjson.dumps(obj, default=default, *args, **kwargs)
    
    @classmethod
    async def _async_decode(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if cls.parser_enabled: return create_simdobj(cls.parser.parse(data, *args, **kwargs))
        return _simdjson.loads(data, *args, **kwargs)



class Json(JsonBase):
    parser: _simdjson.Parser = SimdJson.parser
    parser_enabled: bool = True

    @staticmethod
    def parse(data: Any, *args, **kwargs) -> Union[Union[_simdjson.Object, _simdjson.Array], Union[Dict[Any, Any], List[str]]]:
        return SimdJson.parse(data, *args, **kwargs)
    
    @staticmethod
    def _encode(obj: Any, *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        if isinstance(obj, (SimdObject, SimdArray)): obj = obj.data
        return OrJson._encode(obj, default=default, *args, **kwargs)
    
    @classmethod
    def _decode(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if cls.parser_enabled: return SimdJson.parse(data, *args, **kwargs)
        return OrJson._decode(data, *args, **kwargs)
    
    @staticmethod
    async def _async_encode(obj: Any, *args, default: Dict[Any, Any] = None, **kwargs) -> str:
        if isinstance(obj, (SimdObject, SimdArray)): obj = obj.data
        return await OrJson._async_encode(obj, default=default, *args, **kwargs)
    
    @classmethod
    async def _async_decode(cls, data: Union[str, bytes], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
        if cls.parser_enabled: return SimdJson.parse(data, *args, **kwargs)
        return await OrJson._async_decode(data, *args, **kwargs)


from .base import *
from ._json import JsonSrlzer, OrJson, SimdJson
from ._yaml import Yaml
from ._base import Base


class Srlz:
    Json: JsonSrlzer = OrJson
    Yaml: Yaml = Yaml
    Base: Base = Base
    
    
    



"""
Modified version from fusepy
"""

import sys
import ctypes
import errno
import os
import warnings
import functools
from signal import signal, SIGINT, SIG_DFL
from stat import S_IFDIR
import inspect

try:
    from functools import partial
except ImportError:
    # http://docs.python.org/library/functools.html#functools.partial
    def partial(func, *args, **keywords):
        def newfunc(*fargs, **fkeywords):
            newkeywords = keywords.copy()
            newkeywords.update(fkeywords)
            return func(*(args + fargs), **newkeywords)

        newfunc.func = func
        newfunc.args = args
        newfunc.keywords = keywords
        return newfunc


from typing import Callable, Dict, Any
from types import ModuleType
from lazy.libz import Lib
from lazy.utils import get_logger

logger = get_logger('Fuze')


if Lib.is_avail_fuse:
    import fuse
    _FUZE_READY = True
else: 
    fuse = ModuleType
    _FUZE_READY = False

_FUZE_ALLOWED: bool = not sys.platform.startswith('win')

def _prepare_fuze():
    global _FUZE_READY, fuse
    assert _FUZE_ALLOWED, 'Windows is not supported'
    if _FUZE_READY: return
    
    Lib._ensure_binary_installed('fuse')
    fuse = Lib.import_lib('fuse', 'fusepy')

    _FUZE_READY = True

def async_wrapper(fn):
    @functools.wraps(fn)
    async def wrapper(*args, **kwargs):
        await fn(*args, **kwargs)
    return wrapper


def iscoroutinefunction(obj):
    if inspect.iscoroutinefunction(obj): return True
    if hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__): return True
    return False


class FuseOSError(OSError):
    def __init__(self, errno):
        super(FuseOSError, self).__init__(errno, os.strerror(errno))

class FUZE(object):
    '''
    This class is the lower level interface and should not be subclassed under
    normal use. Its methods are called by fuse.

    Assumes API version 2.6 or later.
    '''

    OPTIONS = (
        ('foreground', '-f'),
        ('debug', '-d'),
        ('nothreads', '-s'),
    )

    def __init__(self, operations, mountpoint, raw_fi=False, encoding='utf-8', **kwargs):

        '''
        Setting raw_fi to True will cause FUSE to pass the fuse_file_info
        class as is to Operations, instead of just the fh field.

        This gives you access to direct_io, keep_cache, etc.
        '''
        _prepare_fuze()
        self.operations = operations
        self.raw_fi = raw_fi
        self.encoding = encoding
        self.__critical_exception = None

        self.use_ns = getattr(operations, 'use_ns', False)
        if not self.use_ns:
            warnings.warn(
                'Time as floating point seconds for utimens is deprecated!\n'
                'To enable time as nanoseconds set the property "use_ns" to '
                'True in your operations class or set your fusepy '
                'requirements to <4.',
                DeprecationWarning)

        args = ['fuse']
        args.extend(flag for arg, flag in self.OPTIONS if kwargs.pop(arg, False))
        kwargs.setdefault('fsname', operations.__class__.__name__)
        args.append('-o')
        args.append(','.join(self._normalize_fuse_options(**kwargs)))
        args.append(mountpoint)

        args = [arg.encode(encoding) for arg in args]
        argv = (ctypes.c_char_p * len(args))(*args)

        fuse_ops = fuse.fuse_operations()
        for ent in fuse.fuse_operations._fields_:
            name, prototype = ent[:2]
            check_name = name
            # ftruncate()/fgetattr() are implemented in terms of their
            # non-f-prefixed versions in the operations object
            if check_name in ["ftruncate", "fgetattr"]: check_name = check_name[1:]

            val = getattr(operations, check_name, None)
            if val is None: continue

            # Function pointer members are tested for using the
            # getattr(operations, name) above but are dynamically
            # invoked using self.operations(name)
            if hasattr(prototype, 'argtypes'):
                val = prototype(partial(self._wrapper, getattr(self, name)))

            setattr(fuse_ops, name, val)

        try: old_handler = signal(SIGINT, SIG_DFL)
        except ValueError: old_handler = SIG_DFL

        err = fuse._libfuse.fuse_main_real(len(args), argv, ctypes.pointer(fuse_ops), ctypes.sizeof(fuse_ops), None)
        try: signal(SIGINT, old_handler)
        except ValueError: pass

        del self.operations     # Invoke the destructor
        if self.__critical_exception: raise self.__critical_exception
        if err: raise RuntimeError(err)

    @staticmethod
    def _normalize_fuse_options(**kargs):
        for key, value in kargs.items():
            if isinstance(value, bool):
                if value is True:
                    yield key
            else:
                yield '%s=%s' % (key, value)

    @staticmethod
    def _wrapper(func, *args, **kwargs):
        'Decorator for the methods that follow'

        try:
            if func.__name__ == "init":
                # init may not fail, as its return code is just stored as
                # private_data field of struct fuse_context
                return func(*args, **kwargs) or 0

            else:
                try:
                    return func(*args, **kwargs) or 0

                except OSError as e:
                    if e.errno > 0:
                        logger.debug("FUSE operation %s raised a %s, returning errno %s.", func.__name__, type(e), e.errno, exc_info=True)
                        return -e.errno
                    else:
                        logger.error("FUSE operation %s raised an OSError with negative errno %s, returning errno.EINVAL.", func.__name__, e.errno, exc_info=True)
                        return -errno.EINVAL

                except Exception:
                    logger.error("Uncaught exception from FUSE operation %s, returning errno.EINVAL.", func.__name__, exc_info=True)
                    return -errno.EINVAL

        except BaseException as e:
            self.__critical_exception = e
            logger.critical("Uncaught critical exception from FUSE operation %s, aborting.", func.__name__, exc_info=True)
            # the raised exception (even SystemExit) will be caught by FUSE
            # potentially causing SIGSEGV, so tell system to stop/interrupt FUSE
            fuse.fuse_exit()
            return -errno.EFAULT

    def _decode_optional_path(self, path):
        # NB: this method is intended for fuse operations that
        #     allow the path argument to be NULL,
        #     *not* as a generic path decoding method
        if path is None: return None
        return path.decode(self.encoding)

    def getattr(self, path, buf):
        return self.fgetattr(path, buf, None)

    def readlink(self, path, buf, bufsize):
        ret = self.operations('readlink', path.decode(self.encoding)).encode(self.encoding)
        # copies a string into the given buffer
        # (null terminated and truncated if necessary)
        data = ctypes.create_string_buffer(ret[:bufsize - 1])
        ctypes.memmove(buf, data, len(data))
        return 0

    def mknod(self, path, mode, dev):
        return self.operations('mknod', path.decode(self.encoding), mode, dev)

    def mkdir(self, path, mode):
        return self.operations('mkdir', path.decode(self.encoding), mode)

    def unlink(self, path):
        return self.operations('unlink', path.decode(self.encoding))

    def rmdir(self, path):
        return self.operations('rmdir', path.decode(self.encoding))

    def symlink(self, source, target):
        'creates a symlink `target -> source` (e.g. ln -s source target)'

        return self.operations('symlink', target.decode(self.encoding), source.decode(self.encoding))

    def rename(self, old, new):
        return self.operations('rename', old.decode(self.encoding), new.decode(self.encoding))

    def link(self, source, target):
        'creates a hard link `target -> source` (e.g. ln source target)'
        return self.operations('link', target.decode(self.encoding), source.decode(self.encoding))

    def chmod(self, path, mode):
        return self.operations('chmod', path.decode(self.encoding), mode)

    def chown(self, path, uid, gid):
        # Check if any of the arguments is a -1 that has overflowed
        if fuse.c_uid_t(uid + 1).value == 0: uid = -1
        if fuse.c_gid_t(gid + 1).value == 0: gid = -1
        return self.operations('chown', path.decode(self.encoding), uid, gid)

    def truncate(self, path, length):
        return self.operations('truncate', path.decode(self.encoding), length)

    def open(self, path, fip):
        fi = fip.contents
        if self.raw_fi: 
            return self.operations('open', path.decode(self.encoding), fi)
        else:
            fi.fh = self.operations('open', path.decode(self.encoding), fi.flags)
            return 0

    def read(self, path, buf, size, offset, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        ret = self.operations('read', self._decode_optional_path(path), size, offset, fh)

        if not ret: return 0

        retsize = len(ret)
        assert retsize <= size, 'actual amount read %d greater than expected %d' % (retsize, size)
        ctypes.memmove(buf, ret, retsize)
        return retsize

    def write(self, path, buf, size, offset, fip):
        data = ctypes.string_at(buf, size)

        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('write', self._decode_optional_path(path), data, offset, fh)

    def statfs(self, path, buf):
        stv = buf.contents
        attrs = self.operations('statfs', path.decode(self.encoding))
        for key, val in attrs.items():
            if hasattr(stv, key): setattr(stv, key, val)
        return 0

    def flush(self, path, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('flush', self._decode_optional_path(path), fh)

    def release(self, path, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('release', self._decode_optional_path(path), fh)

    def fsync(self, path, datasync, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('fsync', self._decode_optional_path(path), datasync, fh)

    def setxattr(self, path, name, value, size, options, *args):
        return self.operations('setxattr', path.decode(self.encoding), name.decode(self.encoding), ctypes.string_at(value, size), options, *args)

    def getxattr(self, path, name, value, size, *args):
        ret = self.operations('getxattr', path.decode(self.encoding), name.decode(self.encoding), *args)
        retsize = len(ret)
        # allow size queries
        if not value: return retsize

        # do not truncate
        if retsize > size: return -errno.ERANGE

        # Does not add trailing 0
        buf = ctypes.create_string_buffer(ret, retsize)
        ctypes.memmove(value, buf, retsize)

        return retsize

    def listxattr(self, path, namebuf, size):
        attrs = self.operations('listxattr', path.decode(self.encoding)) or ''
        ret = '\x00'.join(attrs).encode(self.encoding)
        if len(ret) > 0: ret += '\x00'.encode(self.encoding)

        retsize = len(ret)
        # allow size queries
        if not namebuf: return retsize

        # do not truncate
        if retsize > size: return -errno.ERANGE

        buf = ctypes.create_string_buffer(ret, retsize)
        ctypes.memmove(namebuf, buf, retsize)

        return retsize

    def removexattr(self, path, name):
        return self.operations('removexattr', path.decode(self.encoding), name.decode(self.encoding))

    def opendir(self, path, fip):
        # Ignore raw_fi
        fip.contents.fh = self.operations('opendir', path.decode(self.encoding))
        return 0

    def readdir(self, path, buf, filler, offset, fip):
        # Ignore raw_fi
        for item in self.operations('readdir', self._decode_optional_path(path), fip.contents.fh):
            if isinstance(item, fuse.basestring): name, st, offset = item, None, 0
            else:
                name, attrs, offset = item
                if attrs:
                    st = fuse.c_stat()
                    fuse.set_st_attrs(st, attrs, use_ns=self.use_ns)
                else:
                    st = None

            if filler(buf, name.encode(self.encoding), st, offset) != 0: break

        return 0

    def releasedir(self, path, fip):
        # Ignore raw_fi
        return self.operations('releasedir', self._decode_optional_path(path), fip.contents.fh)

    def fsyncdir(self, path, datasync, fip):
        # Ignore raw_fi
        return self.operations('fsyncdir', self._decode_optional_path(path), datasync, fip.contents.fh)

    def init(self, conn):
        return self.operations('init', '/')

    def destroy(self, private_data):
        return self.operations('destroy', '/')

    def access(self, path, amode):
        return self.operations('access', path.decode(self.encoding), amode)

    def create(self, path, mode, fip):
        fi = fip.contents
        path = path.decode(self.encoding)

        if self.raw_fi: return self.operations('create', path, mode, fi)
        fi.fh = self.operations('create', path, mode)
        return 0

    def ftruncate(self, path, length, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('truncate', self._decode_optional_path(path), length, fh)

    def fgetattr(self, path, buf, fip):
        ctypes.memset(buf, 0, ctypes.sizeof(fuse.c_stat))

        st = buf.contents
        if not fip: fh = fip
        elif self.raw_fi: fh = fip.contents
        else: fh = fip.contents.fh

        attrs = self.operations('getattr', self._decode_optional_path(path), fh)
        fuse.set_st_attrs(st, attrs, use_ns=self.use_ns)
        return 0

    def lock(self, path, fip, cmd, lock):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('lock', self._decode_optional_path(path), fh, cmd, lock)

    def utimens(self, path, buf):
        if buf:
            atime = fuse.time_of_timespec(buf.contents.actime, use_ns=self.use_ns)
            mtime = fuse.time_of_timespec(buf.contents.modtime, use_ns=self.use_ns)
            times = (atime, mtime)
        else:
            times = None

        return self.operations('utimens', path.decode(self.encoding), times)

    def bmap(self, path, blocksize, idx):
        return self.operations('bmap', path.decode(self.encoding), blocksize, idx)

    def ioctl(self, path, cmd, arg, fip, flags, data):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('ioctl', path.decode(self.encoding), cmd, arg, fh, flags, data)


class AsyncFUZE(object):
    '''
    This class is the lower level interface and should not be subclassed under
    normal use. Its methods are called by fuse.

    Assumes API version 2.6 or later.
    '''

    OPTIONS = (
        ('foreground', '-f'),
        ('debug', '-d'),
        ('nothreads', '-s'),
    )

    def __init__(self, operations, mountpoint, raw_fi=False, encoding='utf-8', **kwargs):

        '''
        Setting raw_fi to True will cause FUSE to pass the fuse_file_info
        class as is to Operations, instead of just the fh field.

        This gives you access to direct_io, keep_cache, etc.
        '''
        _prepare_fuze()
        self.operations = operations
        self.raw_fi = raw_fi
        self.encoding = encoding
        self.__critical_exception = None

        self.use_ns = getattr(operations, 'use_ns', False)
        if not self.use_ns:
            warnings.warn(
                'Time as floating point seconds for utimens is deprecated!\n'
                'To enable time as nanoseconds set the property "use_ns" to '
                'True in your operations class or set your fusepy '
                'requirements to <4.',
                DeprecationWarning)

        args = ['fuse']
        args.extend(flag for arg, flag in self.OPTIONS if kwargs.pop(arg, False))
        kwargs.setdefault('fsname', operations.__class__.__name__)
        args.append('-o')
        args.append(','.join(self._normalize_fuse_options(**kwargs)))
        args.append(mountpoint)

        args = [arg.encode(encoding) for arg in args]
        argv = (ctypes.c_char_p * len(args))(*args)

        fuse_ops = fuse.fuse_operations()
        for ent in fuse.fuse_operations._fields_:
            name, prototype = ent[:2]
            check_name = name
            # ftruncate()/fgetattr() are implemented in terms of their
            # non-f-prefixed versions in the operations object
            if check_name in ["ftruncate", "fgetattr"]: check_name = check_name[1:]

            val = getattr(operations, check_name, None)
            if val is None: continue

            # Function pointer members are tested for using the
            # getattr(operations, name) above but are dynamically
            # invoked using self.operations(name)
            if hasattr(prototype, 'argtypes'):
                val = prototype(partial(self._wrapper, getattr(self, name)))

            setattr(fuse_ops, name, val)

        try: old_handler = signal(SIGINT, SIG_DFL)
        except ValueError: old_handler = SIG_DFL

        err = fuse._libfuse.fuse_main_real(len(args), argv, ctypes.pointer(fuse_ops), ctypes.sizeof(fuse_ops), None)
        try: signal(SIGINT, old_handler)
        except ValueError: pass

        del self.operations     # Invoke the destructor
        if self.__critical_exception: raise self.__critical_exception
        if err: raise RuntimeError(err)

    @staticmethod
    def _normalize_fuse_options(**kargs):
        for key, value in kargs.items():
            if isinstance(value, bool):
                if value is True:
                    yield key
            else:
                yield '%s=%s' % (key, value)

    @staticmethod
    async def _wrapper(func, *args, **kwargs):
        'Decorator for the methods that follow'

        try:
            if func.__name__ == "init":
                # init may not fail, as its return code is just stored as
                # private_data field of struct fuse_context
                if iscoroutinefunction(func): return await func(*args, **kwargs) or 0
                return func(*args, **kwargs) or 0

            else:
                try:
                    if iscoroutinefunction(func): return await func(*args, **kwargs) or 0
                    return func(*args, **kwargs) or 0

                except OSError as e:
                    if e.errno > 0:
                        logger.debug("FUSE operation %s raised a %s, returning errno %s.", func.__name__, type(e), e.errno, exc_info=True)
                        return -e.errno
                    else:
                        logger.error("FUSE operation %s raised an OSError with negative errno %s, returning errno.EINVAL.", func.__name__, e.errno, exc_info=True)
                        return -errno.EINVAL

                except Exception:
                    logger.error("Uncaught exception from FUSE operation %s, returning errno.EINVAL.", func.__name__, exc_info=True)
                    return -errno.EINVAL

        except BaseException as e:
            self.__critical_exception = e
            logger.critical("Uncaught critical exception from FUSE operation %s, aborting.", func.__name__, exc_info=True)
            # the raised exception (even SystemExit) will be caught by FUSE
            # potentially causing SIGSEGV, so tell system to stop/interrupt FUSE
            fuse.fuse_exit()
            return -errno.EFAULT

    def _decode_optional_path(self, path):
        # NB: this method is intended for fuse operations that
        #     allow the path argument to be NULL,
        #     *not* as a generic path decoding method
        if path is None: return None
        return path.decode(self.encoding)

    async def getattr(self, path, buf):
        return await self.fgetattr(path, buf, None)

    async def readlink(self, path, buf, bufsize):
        ret = self.operations('readlink', path.decode(self.encoding)).encode(self.encoding)
        # copies a string into the given buffer
        # (null terminated and truncated if necessary)
        data = ctypes.create_string_buffer(ret[:bufsize - 1])
        ctypes.memmove(buf, data, len(data))
        return 0

    async def mknod(self, path, mode, dev):
        return await self.operations('mknod', path.decode(self.encoding), mode, dev)

    async def mkdir(self, path, mode):
        return await self.operations('mkdir', path.decode(self.encoding), mode)

    async def unlink(self, path):
        return await self.operations('unlink', path.decode(self.encoding))

    async def rmdir(self, path):
        return await self.operations('rmdir', path.decode(self.encoding))

    async def symlink(self, source, target):
        'creates a symlink `target -> source` (e.g. ln -s source target)'

        return await self.operations('symlink', target.decode(self.encoding), source.decode(self.encoding))

    async def rename(self, old, new):
        return await self.operations('rename', old.decode(self.encoding), new.decode(self.encoding))

    async def link(self, source, target):
        'creates a hard link `target -> source` (e.g. ln source target)'
        return await self.operations('link', target.decode(self.encoding), source.decode(self.encoding))

    async def chmod(self, path, mode):
        return await self.operations('chmod', path.decode(self.encoding), mode)

    async def chown(self, path, uid, gid):
        # Check if any of the arguments is a -1 that has overflowed
        if fuse.c_uid_t(uid + 1).value == 0: uid = -1
        if fuse.c_gid_t(gid + 1).value == 0: gid = -1
        return await self.operations('chown', path.decode(self.encoding), uid, gid)

    async def truncate(self, path, length):
        return await self.operations('truncate', path.decode(self.encoding), length)

    async def open(self, path, fip):
        fi = fip.contents
        if self.raw_fi: 
            return await self.operations('open', path.decode(self.encoding), fi)
        else:
            fi.fh = await self.operations('open', path.decode(self.encoding), fi.flags)
            return 0

    async def read(self, path, buf, size, offset, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        ret = await self.operations('read', self._decode_optional_path(path), size, offset, fh)

        if not ret: return 0

        retsize = len(ret)
        assert retsize <= size, 'actual amount read %d greater than expected %d' % (retsize, size)
        ctypes.memmove(buf, ret, retsize)
        return retsize

    async def write(self, path, buf, size, offset, fip):
        data = ctypes.string_at(buf, size)

        fh = fip.contents if self.raw_fi else fip.contents.fh
        return await self.operations('write', self._decode_optional_path(path), data, offset, fh)

    async def statfs(self, path, buf):
        stv = buf.contents
        attrs = await self.operations('statfs', path.decode(self.encoding))
        for key, val in attrs.items():
            if hasattr(stv, key): setattr(stv, key, val)
        return 0

    async def flush(self, path, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return await self.operations('flush', self._decode_optional_path(path), fh)

    async def release(self, path, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return await self.operations('release', self._decode_optional_path(path), fh)

    async def fsync(self, path, datasync, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return await self.operations('fsync', self._decode_optional_path(path), datasync, fh)

    async def setxattr(self, path, name, value, size, options, *args):
        return await self.operations('setxattr', path.decode(self.encoding), name.decode(self.encoding), ctypes.string_at(value, size), options, *args)

    async def getxattr(self, path, name, value, size, *args):
        ret = await self.operations('getxattr', path.decode(self.encoding), name.decode(self.encoding), *args)
        retsize = len(ret)
        # allow size queries
        if not value: return retsize

        # do not truncate
        if retsize > size: return -errno.ERANGE

        # Does not add trailing 0
        buf = ctypes.create_string_buffer(ret, retsize)
        ctypes.memmove(value, buf, retsize)

        return retsize

    async def listxattr(self, path, namebuf, size):
        attrs = await self.operations('listxattr', path.decode(self.encoding)) or ''
        ret = '\x00'.join(attrs).encode(self.encoding)
        if len(ret) > 0: ret += '\x00'.encode(self.encoding)

        retsize = len(ret)
        # allow size queries
        if not namebuf: return retsize

        # do not truncate
        if retsize > size: return -errno.ERANGE

        buf = ctypes.create_string_buffer(ret, retsize)
        ctypes.memmove(namebuf, buf, retsize)

        return retsize

    async def removexattr(self, path, name):
        return await self.operations('removexattr', path.decode(self.encoding), name.decode(self.encoding))

    async def opendir(self, path, fip):
        # Ignore raw_fi
        fip.contents.fh = await self.operations('opendir', path.decode(self.encoding))
        return 0

    async def readdir(self, path, buf, filler, offset, fip):
        # Ignore raw_fi
        for item in await self.operations('readdir', self._decode_optional_path(path), fip.contents.fh):
            if isinstance(item, fuse.basestring): name, st, offset = item, None, 0
            else:
                name, attrs, offset = item
                if attrs:
                    st = fuse.c_stat()
                    fuse.set_st_attrs(st, attrs, use_ns=self.use_ns)
                else:
                    st = None

            if filler(buf, name.encode(self.encoding), st, offset) != 0: break

        return 0

    async def releasedir(self, path, fip):
        # Ignore raw_fi
        return await self.operations('releasedir', self._decode_optional_path(path), fip.contents.fh)

    async def fsyncdir(self, path, datasync, fip):
        # Ignore raw_fi
        return await self.operations('fsyncdir', self._decode_optional_path(path), datasync, fip.contents.fh)

    async def init(self, conn):
        return await self.operations('init', '/')

    async def destroy(self, private_data):
        return await self.operations('destroy', '/')

    async def access(self, path, amode):
        return await self.operations('access', path.decode(self.encoding), amode)

    async def create(self, path, mode, fip):
        fi = fip.contents
        path = path.decode(self.encoding)

        if self.raw_fi: return await self.operations('create', path, mode, fi)
        fi.fh = await self.operations('create', path, mode)
        return 0

    async def ftruncate(self, path, length, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return await self.operations('truncate', self._decode_optional_path(path), length, fh)

    async def fgetattr(self, path, buf, fip):
        ctypes.memset(buf, 0, ctypes.sizeof(fuse.c_stat))

        st = buf.contents
        if not fip: fh = fip
        elif self.raw_fi: fh = fip.contents
        else: fh = fip.contents.fh

        attrs = await self.operations('getattr', self._decode_optional_path(path), fh)
        fuse.set_st_attrs(st, attrs, use_ns=self.use_ns)
        return 0

    async def lock(self, path, fip, cmd, lock):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return await self.operations('lock', self._decode_optional_path(path), fh, cmd, lock)

    async def utimens(self, path, buf):
        if buf:
            atime = fuse.time_of_timespec(buf.contents.actime, use_ns=self.use_ns)
            mtime = fuse.time_of_timespec(buf.contents.modtime, use_ns=self.use_ns)
            times = (atime, mtime)
        else:
            times = None

        return await self.operations('utimens', path.decode(self.encoding), times)

    async def bmap(self, path, blocksize, idx):
        return await self.operations('bmap', path.decode(self.encoding), blocksize, idx)

    async def ioctl(self, path, cmd, arg, fip, flags, data):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return await self.operations('ioctl', path.decode(self.encoding), cmd, arg, fh, flags, data)
    
    


class Operations(object):
    '''
    This class should be subclassed and passed as an argument to FUSE on
    initialization. All operations should raise a FuseOSError exception on
    error.

    When in doubt of what an operation should do, check the FUSE header file
    or the corresponding system call man page.
    '''

    def __call__(self, op, *args):
        if not hasattr(self, op):
            raise FuseOSError(errno.EFAULT)
        return getattr(self, op)(*args)

    def access(self, path, amode):
        return 0

    bmap = None

    def chmod(self, path, mode):
        raise FuseOSError(errno.EROFS)

    def chown(self, path, uid, gid):
        raise FuseOSError(errno.EROFS)

    def create(self, path, mode, fi=None):
        '''
        When raw_fi is False (default case), fi is None and create should
        return a numerical file handle.

        When raw_fi is True the file handle should be set directly by create
        and return 0.
        '''

        raise FuseOSError(errno.EROFS)

    def destroy(self, path):
        'Called on filesystem destruction. Path is always /'

        pass

    def flush(self, path, fh):
        return 0

    def fsync(self, path, datasync, fh):
        return 0

    def fsyncdir(self, path, datasync, fh):
        return 0

    def getattr(self, path, fh=None):
        '''
        Returns a dictionary with keys identical to the stat C structure of
        stat(2).

        st_atime, st_mtime and st_ctime should be floats.

        NOTE: There is an incompatibility between Linux and Mac OS X
        concerning st_nlink of directories. Mac OS X counts all files inside
        the directory, while Linux counts only the subdirectories.
        '''

        if path != '/':
            raise FuseOSError(errno.ENOENT)
        return dict(st_mode=(S_IFDIR | 0o755), st_nlink=2)

    def getxattr(self, path, name, position=0):
        raise FuseOSError(fuse.ENOTSUP)

    def init(self, path):
        '''
        Called on filesystem initialization. (Path is always /)

        Use it instead of __init__ if you start threads on initialization.
        '''

        pass

    def ioctl(self, path, cmd, arg, fip, flags, data):
        raise FuseOSError(errno.ENOTTY)

    def link(self, target, source):
        'creates a hard link `target -> source` (e.g. ln source target)'

        raise FuseOSError(errno.EROFS)

    def listxattr(self, path):
        return []

    lock = None

    def mkdir(self, path, mode):
        raise FuseOSError(errno.EROFS)

    def mknod(self, path, mode, dev):
        raise FuseOSError(errno.EROFS)

    def open(self, path, flags):
        '''
        When raw_fi is False (default case), open should return a numerical
        file handle.

        When raw_fi is True the signature of open becomes:
            open(self, path, fi)

        and the file handle should be set directly.
        '''

        return 0

    def opendir(self, path):
        'Returns a numerical file handle.'

        return 0

    def read(self, path, size, offset, fh):
        'Returns a string containing the data requested.'

        raise FuseOSError(errno.EIO)

    def readdir(self, path, fh):
        '''
        Can return either a list of names, or a list of (name, attrs, offset)
        tuples. attrs is a dict as in getattr.
        '''

        return ['.', '..']

    def readlink(self, path):
        raise FuseOSError(errno.ENOENT)

    def release(self, path, fh):
        return 0

    def releasedir(self, path, fh):
        return 0

    def removexattr(self, path, name):
        raise FuseOSError(fuse.ENOTSUP)

    def rename(self, old, new):
        raise FuseOSError(errno.EROFS)

    def rmdir(self, path):
        raise FuseOSError(errno.EROFS)

    def setxattr(self, path, name, value, options, position=0):
        raise FuseOSError(fuse.ENOTSUP)

    def statfs(self, path):
        '''
        Returns a dictionary with keys identical to the statvfs C structure of
        statvfs(3).

        On Mac OS X f_bsize and f_frsize must be a power of 2
        (minimum 512).
        '''

        return {}

    def symlink(self, target, source):
        'creates a symlink `target -> source` (e.g. ln -s source target)'

        raise FuseOSError(errno.EROFS)

    def truncate(self, path, length, fh=None):
        raise FuseOSError(errno.EROFS)

    def unlink(self, path):
        raise FuseOSError(errno.EROFS)

    def utimens(self, path, times=None):
        'Times is a (atime, mtime) tuple. If None use current time.'

        return 0

    def write(self, path, data, offset, fh):
        raise FuseOSError(errno.EROFS)

class AsyncOperations(object):
    '''
    This class should be subclassed and passed as an argument to FUSE on
    initialization. All operations should raise a FuseOSError exception on
    error.

    When in doubt of what an operation should do, check the FUSE header file
    or the corresponding system call man page.
    '''

    async def __call__(self, op, *args):
        if not hasattr(self, op):
            raise FuseOSError(errno.EFAULT)
        a = getattr(self, op)
        if iscoroutinefunction(a): return a(*args)
        return await a(*args)

    async def access(self, path, amode):
        return 0

    bmap = None

    async def chmod(self, path, mode):
        raise FuseOSError(errno.EROFS)

    async def chown(self, path, uid, gid):
        raise FuseOSError(errno.EROFS)

    async def create(self, path, mode, fi=None):
        '''
        When raw_fi is False (default case), fi is None and create should
        return a numerical file handle.

        When raw_fi is True the file handle should be set directly by create
        and return 0.
        '''

        raise FuseOSError(errno.EROFS)

    async def destroy(self, path):
        'Called on filesystem destruction. Path is always /'

        pass

    async def flush(self, path, fh):
        return 0

    async def fsync(self, path, datasync, fh):
        return 0

    async def fsyncdir(self, path, datasync, fh):
        return 0

    async def getattr(self, path, fh=None):
        '''
        Returns a dictionary with keys identical to the stat C structure of
        stat(2).

        st_atime, st_mtime and st_ctime should be floats.

        NOTE: There is an incompatibility between Linux and Mac OS X
        concerning st_nlink of directories. Mac OS X counts all files inside
        the directory, while Linux counts only the subdirectories.
        '''

        if path != '/':
            raise FuseOSError(errno.ENOENT)
        return dict(st_mode=(S_IFDIR | 0o755), st_nlink=2)

    async def getxattr(self, path, name, position=0):
        raise FuseOSError(fuse.ENOTSUP)

    async def init(self, path):
        '''
        Called on filesystem initialization. (Path is always /)

        Use it instead of __init__ if you start threads on initialization.
        '''

        pass

    async def ioctl(self, path, cmd, arg, fip, flags, data):
        raise FuseOSError(errno.ENOTTY)

    async def link(self, target, source):
        'creates a hard link `target -> source` (e.g. ln source target)'

        raise FuseOSError(errno.EROFS)

    async def listxattr(self, path):
        return []

    lock = None

    async def mkdir(self, path, mode):
        raise FuseOSError(errno.EROFS)

    async def mknod(self, path, mode, dev):
        raise FuseOSError(errno.EROFS)

    async def open(self, path, flags):
        '''
        When raw_fi is False (default case), open should return a numerical
        file handle.

        When raw_fi is True the signature of open becomes:
            open(self, path, fi)

        and the file handle should be set directly.
        '''

        return 0

    async def opendir(self, path):
        'Returns a numerical file handle.'

        return 0

    async def read(self, path, size, offset, fh):
        'Returns a string containing the data requested.'

        raise FuseOSError(errno.EIO)

    async def readdir(self, path, fh):
        '''
        Can return either a list of names, or a list of (name, attrs, offset)
        tuples. attrs is a dict as in getattr.
        '''

        return ['.', '..']

    async def readlink(self, path):
        raise FuseOSError(errno.ENOENT)

    async def release(self, path, fh):
        return 0

    async def releasedir(self, path, fh):
        return 0

    async def removexattr(self, path, name):
        raise FuseOSError(fuse.ENOTSUP)

    async def rename(self, old, new):
        raise FuseOSError(errno.EROFS)

    async def rmdir(self, path):
        raise FuseOSError(errno.EROFS)

    async def setxattr(self, path, name, value, options, position=0):
        raise FuseOSError(fuse.ENOTSUP)

    async def statfs(self, path):
        '''
        Returns a dictionary with keys identical to the statvfs C structure of
        statvfs(3).

        On Mac OS X f_bsize and f_frsize must be a power of 2
        (minimum 512).
        '''

        return {}

    async def symlink(self, target, source):
        'creates a symlink `target -> source` (e.g. ln -s source target)'

        raise FuseOSError(errno.EROFS)

    async def truncate(self, path, length, fh=None):
        raise FuseOSError(errno.EROFS)

    async def unlink(self, path):
        raise FuseOSError(errno.EROFS)

    async def utimens(self, path, times=None):
        'Times is a (atime, mtime) tuple. If None use current time.'

        return 0

    async def write(self, path, data, offset, fh):
        raise FuseOSError(errno.EROFS)



from typing import List, Dict, Optional, Callable, TypeVar, Any, Union, Coroutine, Generator, Type, Tuple

Data = TypeVar('Data', str, List[str], Dict[str, Union[str, List[str]]])
AnyMany = TypeVar('AnyMany', Any, List[Any])

TextMany = TypeVar('TextMany', str, List[str])
TextList = List[str]

DictList = List[Dict[str, Any]]
DictMany = TypeVar('DictMany', Dict[str, Any], List[Dict[str, Any]])
DictAny = Dict[str, Any]
DictText = Dict[str, str]

DefaultHeaders = {
    'Accept': 'application/json',
    'Content-Type': 'application/json'
}



import os
import ntpath
import pathlib
import posixpath

from .core import PosixFSxPath

"""
Type Checking
"""

try: import s3fs
except ImportError: s3fs = object

try: import gcsfs
except ImportError: gcsfs = object

try: from lazy import CloudAuthz
except ImportError: CloudAuthz = object

_authz: CloudAuthz = None

def get_cloudauthz():
    global _authz
    if _authz is None:
        from lazy import CloudAuthz
        _authz = CloudAuthz
    return _authz


class PosixS3Path(PosixFSxPath, pathlib.PurePosixPath):
    """
    AWS S3 Pathlib-like API around `fsspec.s3fs` providing Async Capabilities
    """
    _PATH = posixpath
    _FSX: 's3fs.S3FileSystem' = None
    _SYNC_FS: 's3fs.S3FileSystem' = None
    _ASYNC_FS: 's3fs.S3FileSystem' = None
    _FSX_LIB: str = 's3fs'
    _FSX_CLS: str = 'S3FileSystem'
    _FSX_MODULE = None
    _AUTHZ: 'CloudAuthz' = None

    @property
    def async_fs(self) -> 's3fs.S3FileSystem':
        if not self._ASYNC_FS: self.get_filesystem(is_async=True)
        return self._ASYNC_FS
    
    @property
    def sync_fs(self) -> 's3fs.S3FileSystem':
        if not self._SYNC_FS: self.get_filesystem()
        return self._SYNC_FS
    
    @classmethod
    def get_authz(cls, reload: bool = False, **config):
        if cls._AUTHZ and not reload: return cls._AUTHZ
        cls._AUTHZ = get_cloudauthz()
        if config: cls._AUTHZ.update_authz(**config)
        return cls._AUTHZ

    @classmethod
    def get_configz(cls, reload: bool = False, **config):
        authz = cls.get_authz(reload=reload, **config)
        _config = {}
        if authz.aws_access_key_id:
            _config['key'] = authz.aws_access_key_id
            _config['secret'] = authz.aws_secret_access_key
        elif authz.aws_access_token: _config['token'] = authz.aws_access_token
        elif not authz.boto_config: _config['anon'] = True
        if authz.s3_config: _config['config_kwargs'] = authz.s3_config
        return _config
    

class PosixGCSPath(PosixFSxPath, pathlib.PurePosixPath):
    """
    Google Cloud Storage Pathlib-like API around `fsspec.gcsfs` providing Async Capabilities
    """
    _PATH = posixpath
    _FSX: 'gcsfs.GCSFileSystem' = None
    _SYNC_FS: 'gcsfs.GCSFileSystem' = None
    _ASYNC_FS: 'gcsfs.GCSFileSystem' = None
    _FSX_LIB: str = 'gcsfs'
    _FSX_CLS: str = 'GCSFileSystem'
    _FSX_MODULE = None
    _AUTHZ: 'CloudAuthz' = None

    @classmethod
    def get_authz(cls, reload: bool = False, **config):
        if cls._AUTHZ and not reload: return cls._AUTHZ
        cls._AUTHZ = get_cloudauthz()
        if config: cls._AUTHZ.update_authz(**config)
        return cls._AUTHZ

    @classmethod
    def get_configz(cls, reload: bool = False, **config):
        authz = cls.get_authz(reload=reload, **config)
        _config = {}
        if authz.gcp_auth: _config['token'] = authz.gcp_auth
        if authz.gcloud_project or authz.google_cloud_project: _config['project'] = authz.gcloud_project or authz.google_cloud_project
        if authz.gcs_client_config: _config['client_kwargs'] = authz.gcs_client_config
        if authz.gcs_config: _config['config_kwargs'] = authz.gcs_config
        return _config

    @property
    def async_fs(self) -> 'gcsfs.GCSFileSystem':
        if not self._ASYNC_FS: self.get_filesystem(is_async=True)
        return self._ASYNC_FS
    
    @property
    def sync_fs(self) -> 'gcsfs.GCSFileSystem':
        if not self._SYNC_FS: self.get_filesystem()
        return self._SYNC_FS


class PosixMinioPath(PosixS3Path, pathlib.PurePosixPath):
    """
    Minio-S3 Pathlib-like API around `fsspec.s3fs` providing Async Capabilities
    """
    _PATH = posixpath
    _FSX: 's3fs.S3FileSystem' = None
    _SYNC_FS: 's3fs.S3FileSystem' = None
    _ASYNC_FS: 's3fs.S3FileSystem' = None
    _FSX_LIB: str = 's3fs'
    _FSX_CLS: str = 'S3FileSystem'
    _FSX_MODULE = None
    _AUTHZ: 'CloudAuthz' = None

    @classmethod
    def get_configz(cls, reload: bool = False, **config):
        authz = cls.get_authz(reload=reload, **config)
        _config = {}
        if authz.minio_secret_key:
            _config['key'] = authz.minio_access_key
            _config['secret'] = authz.minio_secret_key
        elif authz.minio_access_token: _config['token'] = authz.minio_access_token
        _config['client_kwargs'] = {'endpoint_url': authz.minio_endpoint}
        if authz.minio_config: _config['config_kwargs'] = authz.minio_config
        return _config
    

class PosixS3CompatPath(PosixS3Path, pathlib.PurePosixPath):
    """
    S3-Compatible Pathlib-like API around `fsspec.s3fs` providing Async Capabilities
    """
    _PATH = posixpath
    _FSX: 's3fs.S3FileSystem' = None
    _SYNC_FS: 's3fs.S3FileSystem' = None
    _ASYNC_FS: 's3fs.S3FileSystem' = None
    _FSX_LIB: str = 's3fs'
    _FSX_CLS: str = 'S3FileSystem'
    _FSX_MODULE = None
    _AUTHZ: 'CloudAuthz' = None

    @classmethod
    def get_configz(cls, reload: bool = False, **config):
        authz = cls.get_authz(reload=reload, **config)
        _config = {}
        if authz.s3compat_secret_key:
            _config['key'] = authz.s3compat_access_key
            _config['secret'] = authz.s3compat_secret_key
        elif authz.s3compat_access_token: _config['token'] = authz.s3compat_access_token
        _config['client_kwargs'] = {'endpoint_url': authz.s3compat_endpoint}
        if authz.s3compat_region: _config['client_kwargs']['region_name'] = authz.s3compat_region
        if authz.s3compat_config: _config['config_kwargs'] = authz.s3compat_config
        return _config



class WindowsGCSPath(PosixGCSPath, pathlib.PureWindowsPath):
    _PATH = ntpath

class WindowsS3Path(PosixS3Path, pathlib.PureWindowsPath):
    _PATH = ntpath

class WindowsMinioPath(PosixMinioPath, pathlib.PureWindowsPath):
    _PATH = ntpath

class WindowsS3CompatPath(PosixS3CompatPath, pathlib.PureWindowsPath):
    _PATH = ntpath


os.PathLike.register(PosixGCSPath)
os.PathLike.register(PosixS3Path)
os.PathLike.register(PosixMinioPath)
os.PathLike.register(PosixS3CompatPath)

os.PathLike.register(WindowsGCSPath)
os.PathLike.register(WindowsS3Path)
os.PathLike.register(WindowsMinioPath)
os.PathLike.register(WindowsS3CompatPath)




import abc
import os
import typing

from lazy.serialize import Serialize
from .types import *

from logz import get_logger
logger = get_logger('lazyio')

__all__ = (
    'DEFAULT_ENCODING',
    'PurePath',
    'ReadOnlyPath',
    'ReadWritePath'
    'Serialize',
    'logger'
)

DEFAULT_ENCODING = 'utf-8'


class PurePath(Protocol):
    """Protocol for pathlib.PurePath-like API."""
    parts: Tuple[str, ...]
    drive: str
    root: str
    anchor: str
    name: str
    suffix: str
    suffixes: List[str]
    stem: str

    def __new__(cls: Type[T], *args: PathLike) -> T:
        raise NotImplementedError

    def __fspath__(self) -> str:
        raise NotImplementedError

    def __hash__(self) -> int:
        raise NotImplementedError

    def __lt__(self, other: 'PurePath') -> bool:
        raise NotImplementedError

    def __le__(self, other: 'PurePath') -> bool:
        raise NotImplementedError

    def __gt__(self, other: 'PurePath') -> bool:
        raise NotImplementedError

    def __ge__(self, other: 'PurePath') -> bool:
        raise NotImplementedError

    def __truediv__(self: T, key: PathLike) -> T:
        raise NotImplementedError

    def __rtruediv__(self: T, key: PathLike) -> T:
        raise NotImplementedError

    def __bytes__(self) -> bytes:
        raise NotImplementedError

    def as_posix(self) -> str:
        raise NotImplementedError

    def as_uri(self) -> str:
        raise NotImplementedError

    def is_absolute(self) -> bool:
        raise NotImplementedError

    def is_reserved(self) -> bool:
        raise NotImplementedError

    def match(self, path_pattern: str) -> bool:
        raise NotImplementedError

    def relative_to(self: T, *other: PathLike) -> T:
        raise NotImplementedError

    def with_name(self: T, name: str) -> T:
        raise NotImplementedError

    def with_suffix(self: T, suffix: str) -> T:
        raise NotImplementedError

    def joinpath(self: T, *other: PathLike) -> T:
        raise NotImplementedError
    
    def get_blob(self: T) -> T:
        raise NotImplementedError

    @property
    def parents(self: T) -> Sequence[T]:
        raise NotImplementedError

    @property
    def parent(self: T) -> T:
        raise NotImplementedError
    
    @property
    def absolute_parent(self: T) -> T:
        raise NotImplementedError

    @property
    def is_cloud(self: T) -> bool:
        raise NotImplementedError
    
    @property
    def cloud_provider(self: T) -> T:
        raise NotImplementedError

    @property
    def is_gs(self: T) -> bool:
        raise NotImplementedError
    
    @property
    def is_s3(self: T) -> bool:
        raise NotImplementedError
     
    @property
    def is_minio(self: T) -> bool:
        raise NotImplementedError
    
    @property
    def file_ext(self: T) -> T:
        return self.suffix[1:]

    @property
    def extension(self: T) -> T:
        return self.suffix[1:]

    @property
    def bucket(self: T) -> Optional[str]:
        raise NotImplementedError

    @property
    def bucket_path(self) -> Optional[str]:
        raise NotImplementedError

    # py3.9 backport of PurePath.is_relative_to.
    def is_relative_to(self, *other: PathLike) -> bool:
        """Return True if the path is relative to another path or False."""
        try:
            self.relative_to(*other)
            return True
        except ValueError: return False


class ReadOnlyPath(PurePath, Protocol):
    """Protocol for read-only methods of pathlib.Path-like API.
    See [pathlib.Path](https://docs.python.org/3/library/pathlib.html)
    documentation.
    """

    def __new__(cls: Type[T], *args: PathLike) -> T:
        if cls not in (ReadOnlyPath, ReadWritePath):
            return super().__new__(cls, *args)
        from .generic import as_path
        return as_path(*args)

    @abc.abstractmethod
    def exists(self) -> bool:
        """Returns True if self exists."""

    @abc.abstractmethod
    def is_dir(self) -> bool:
        """Returns True if self is a dir."""

    def is_file(self) -> bool:
        """Returns True if self is a file."""
        return not self.is_dir()

    @abc.abstractmethod
    def iterdir(self: T) -> Iterator[T]:
        """Iterates over the directory."""

    @abc.abstractmethod
    def glob(self: T, pattern: str) -> Iterator[T]:
        """Yielding all matching files (of any kind)."""
        # Might be able to implement using `iterdir` (recursivelly for `rglob`).
    
    def get_files(self: T, pattern: str) -> List[T]:
        """returns all found results from glob"""
        it = self.glob(pattern)
        return [i for i in it if i.is_file()]

    def rglob(self: T, pattern: str) -> Iterator[T]:
        """Yielding all matching files recursivelly (of any kind)."""
        return self.glob(f'**/{pattern}')

    def expanduser(self: T) -> T:
        """Returns a new path with expanded `~` and `~user` constructs."""
        if '~' not in self.parts:    # pytype: disable=attribute-error
            return self
        raise NotImplementedError

    @abc.abstractmethod
    def resolve(self: T, strict: bool = False) -> T:
        """Returns the absolute path."""

    @abc.abstractmethod
    def open(
            self,
            mode: str = 'r',
            encoding: Optional[str] = DEFAULT_ENCODING,
            errors: Optional[str] = None,
            **kwargs: Any,
    ) -> typing.IO[AnyStr]:
        """Opens the file."""
    
    def read(self, mode='r', encoding: Optional[str] = DEFAULT_ENCODING, **kwargs) -> str:
        """ Reads the File in r """
        with self.open(mode=mode, encoding=encoding, **kwargs) as f:
            return f.read()
    
    def read_b(self, mode='rb', **kwargs) -> bytes:
        """ Reads the File in rb """
        with self.open(mode=mode) as f:
            return f.read()
    
    def readlines(self, **kwargs) -> List[str]:
        with self.open('r', **kwargs) as f:
            return f.readlines()

    def read_bytes(self) -> bytes:
        """Reads contents of self as bytes."""
        with self.open('rb') as f:
            return f.read()

    def read_text(self, encoding: Optional[str] = DEFAULT_ENCODING, **kwargs) -> str:
        """Reads contents of self as bytes."""
        with self.open('r', encoding=encoding, **kwargs) as f:
            return f.read()

    def format(self: T, *args: Any, **kwargs: Any) -> T:
        """Apply `str.format()` to the path."""
        return type(self)(os.fspath(self).format(*args, **kwargs))    # pytype: disable=not-instantiable

    
    """ Serializer Specific Methods """

    def read_json(self, encoding: Optional[str] = DEFAULT_ENCODING, **kwargs) -> JsonType:
        #with self.open('r', encoding=encoding) as f:
        return Serialize.Json.loads(self.read_text(encoding = encoding), **kwargs)
        
    def read_jsonlines(self,  mode: str = 'r', skip_errors: bool = True, as_iterable: bool = True, **kwargs) -> Iterator[T]:
        with self.open(mode=mode) as f:
            return Serialize.Json.readlines(f, as_iterable = as_iterable, skip_errors = skip_errors, **kwargs)
    
    def read_yaml(self, encoding: Optional[str] = DEFAULT_ENCODING, **kwargs) -> JsonType:
        #with self.open('r', encoding=encoding) as f:
        return Serialize.Yaml.loads(self.read_text(encoding = encoding), **kwargs)
    
    def read_pickle(self, mode: str = 'rb', **kwargs):
        with self.open(mode=mode) as f:
            return Serialize.Pkl.loads(f.read(), **kwargs)
    


class ReadWritePath(ReadOnlyPath, Protocol):
    """Protocol for pathlib.Path-like API.
    See [pathlib.Path](https://docs.python.org/3/library/pathlib.html)
    documentation.
    """

    @abc.abstractmethod
    def mkdir(self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False) -> None:
        """Create a new directory at this given path."""

    @abc.abstractmethod
    def rmdir(self) -> None:
        """Remove the empty directory at this given path."""

    @abc.abstractmethod
    def rmtree(self) -> None:
        """Remove the directory, including all sub-files."""

    @abc.abstractmethod
    def unlink(self, missing_ok: bool = False) -> None:
        """Remove this file or symbolic link."""
    
    def ensure_dir(self: T, mode: int = 0o777, parents: bool = True, exist_ok: bool = True):
        """Ensures the parent directory exists, creates if not"""
        return self.absolute_parent.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)

    def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
        """Create a file at this given path."""
        del mode    # Unused
        if self.exists():
            if exist_ok: return
            else: raise FileExistsError(f'{self} already exists.')
        self.write_text('', mode='w')

    @abc.abstractmethod
    def rename(self: T, target: PathLike) -> T:
        """Renames the path."""

    @abc.abstractmethod
    def replace(self: T, target: PathLike) -> T:
        """Overwrites the destination path."""

    @abc.abstractmethod
    def copy(self: T, dst: PathLike, overwrite: bool = False) -> T:
        """Copy the current file to the given destination."""
    
    @property
    def absolute_parent(self: T) -> T:
        uri_scheme = self._uri_scheme
        if uri_scheme: return self._new(self._PATH.join(f'{uri_scheme}://', '/'.join(self.parts[2:-1])))
        p = self.resolve()
        if p.is_dir:  return p
        return p.parent


    def copydir(self: T, dst: PathLike, ignore=['.git'], overwrite: bool = False, dryrun: bool = False, **kwargs) -> List[T]:
        """Copies the Current Top Level Parent Dir to the Dst Dir without recursion"""
        dst = self._new(dst)
        assert dst.is_dir(), 'Destination is not a valid directory'
        if not dryrun: dst.ensure_dir()
        copied_files = []
        fnames = self.listdir(ignore=ignore)
        curdir = self.absolute_parent
        for fname in fnames:
            dest_path = dst.joinpath(fname.relative_to(curdir))
            if not dryrun: fname.copy(dest_path, overwrite=overwrite, skip_errors=True)
            copied_files.append(dest_path)
        return copied_files


    def copydirs(self: T, dst: PathLike, mode: str = 'shallow', pattern='*', ignore=['.git'], overwrite: bool = False, levels: int = 2, dryrun: bool = False, **kwargs) -> List[T]:
        """Copies the Current Parent Dir to the Dst Dir.
        modes = [shallow for top level recursive. recursive for all nested]
        levels = number of recursive levels
        dryrun = returns all files that would have been copied without copying
        """
        assert mode in {'shallow', 'recursive'}, 'Invalid Mode Option: [shallow, recursive]'
        dst = self._new(dst)
        assert dst.is_dir(), 'Destination is not a valid directory'
        levels = max(1, levels)
        dst.ensure_dir()
        curdir = self.absolute_parent
        copied_files = []
        if levels > 1 and mode == 'recursive' and '/' not in pattern:
            for _ in range(1, levels): pattern += '/*'
        if self.is_dir() and not pattern.startswith('/'): pattern = '*/' + pattern
        fiter = curdir.glob(pattern) if mode == 'shallow' else curdir.rglob(pattern)
        fnames = [f for f in fiter if not bool(set(f.parts).intersection(ignore))]
        for f in fnames:
            dest_path = dst.joinpath(f.relative_to(curdir))
            if not dryrun:
                if f.is_dir(): dest_path.ensure_dir()
                else: f.copy(dest_path, overwrite=overwrite, skip_errors=True)
            copied_files.append(dest_path)
        return copied_files

    def write_bytes(self, data: bytes) -> None:
        """Writes content as bytes."""
        with self.open('wb') as f:
            return f.write(data)

    def write_text(self, data: str, append: bool = False, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = None) -> None:
        """Writes content as str."""
        mode = 'a' if append and self.exists() else 'w'
        with self.open(mode, encoding=encoding, errors=errors) as f:
            if 'a' in mode:
                f.write('\n')
            return f.write(data)
    

    """ Serializer Specific Methods """

    def write_pickle(self, obj: Any, **kwargs) -> None:
        data = Serialize.Pkl.dumps(obj, **kwargs)
        return self.write_bytes(data)

    def write_yaml(self, data: JsonType, encoding: Optional[str] = DEFAULT_ENCODING, **kwargs) -> None:
        """Writes Json to File"""
        with self.open('w', encoding = encoding) as f:
            f.write(Serialize.Yaml.dumps(data, **kwargs))

    def write_json(self, data: JsonType, encoding: Optional[str] = DEFAULT_ENCODING, ensure_ascii: bool = False, indent: int = 2, **kwargs) -> None:
        """Writes Json to File"""
        with self.open('w', encoding = encoding) as f:
            f.write(Serialize.Json.dumps(data, ensure_ascii=ensure_ascii, indent=indent, **kwargs))
    
    def write_jsonlines(self, data: List[JsonType], append: bool = False, encoding: Optional[str] = DEFAULT_ENCODING, newline: str = '\n', ignore_errors: bool = True, ensure_file_exists: bool = True, flush_every: int = 0, log_errors: bool = False, **kwargs):
        if ensure_file_exists and not self.exists(): self.touch()
        mode = 'a' if (append and self.exists()) or ensure_file_exists else 'w'
        with self.open(mode=mode, encoding=encoding) as f:
            for n, i in enumerate(data):
                try:
                    d = Serialize.Json.dumps(i, **kwargs)
                    f.write(d + newline)
                    if flush_every and n + 1 % flush_every == 0:
                        f.flush()
                except (StopIteration, KeyboardInterrupt, GeneratorExit): break
                except ValueError as e:
                    if log_errors: logger.error(f'Value Error on idx {n}:\nError: {e}\nItem: {i}')
                    if ignore_errors: continue
                    raise e
                except Exception as e:
                    if log_errors: logger.error(f'Error on idx {n}:\nError: {e}\nItem: {i}')
                    if ignore_errors: continue
                    raise e

    



"""
Extended Classes based off of 
https://github.com/kbairak/pipepy/blob/master/src/pipepy/misc.py
"""

import os
import re
from .base import *

class cd:
    """ `cd` replacement that can be also used as a context processor.
        Equivalent to `os.chdir`:
            >>> from lazycls.ext._cmd.contrib import cd, pwd
            >>> print(pwd())
            <<< /foo
            >>> cd('bar')
            >>> print(pwd())
            <<< /foo/bar
            >>> cd('..')
            >>> print(pwd())
            <<< /foo
        Usage as context processor
            >>> from lazycls.ext._cmd.contrib import cd, pwd
            >>> print(pwd())
            <<< /foo
            >>> with cd('bar'):
            ...     print(pwd())
            <<< /foo/bar
            >>> print(pwd())
            <<< /foo
    """

    def __init__(self, *args, **kwargs):
        self._previous_dir = os.path.abspath(os.curdir)
        os.chdir(*args, **kwargs)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(self._previous_dir)


class export:
    """ `export` replacement that can also be used as a context processor.
        Equivalent to `export`:
            >>> import os
            >>> from lazycls.ext._cmd.contrib import export
            >>> print(os.environ['PATH'])
            <<< foo
            >>> export(PATH="foo:bar")
            >>> print(os.environ['PATH'])
            <<< foo:bar
            >>> export(PATH="foo")
            >>> print(os.environ['PATH'])
            <<< foo
        Usage as a context processor:
            >>> import os
            >>> from lazycls.ext._cmd.contrib import export
            >>> print(os.environ['PATH'])
            <<< foo
            >>> with export(PATH="foo:bar"):
            ...     print(os.environ['PATH'])
            <<< foo:bar
            >>> print(os.environ['PATH'])
            <<< foo
        If an env variable is further changed within the body of `with`, it is
        not restored.
            >>> with export(PATH="foo:bar"):
            ...     export(PATH="foo:BAR")
            >>> print(os.environ['PATH'])
            <<< foo:BAR
    """

    def __init__(self, **kwargs):
        self._previous_env = dict(os.environ)
        self._kwargs = kwargs
        os.environ.update(kwargs)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        for key, value in self._kwargs.items():
            if os.environ[key] != value:
                # Variable changed within the body of the `with` block, skip
                pass
            elif key in self._previous_env:
                # Value was changed by the `with` statement, restore
                os.environ[key] = self._previous_env[key]
            else:
                # Value was added by the `with` statement, delete
                del os.environ[key]


def source(filename, *, recursive=False, quiet=True, shell="bash"):
    """ Source a bash script and export any environment variables defined
        there.
        - filename: The name of the file being sourced, defaults to 'env'
        - recursive: Whether to go through all the parent directories to find
              similarly named bash scripts, defaults to `False`
        - shell: which shell to use for sourcing, defaults to 'bash'
        Can also be used as a context processor for temporary environment
        changes, like `export` (in fact, it uses `export` internally).
        Usage:
        Assuming our directory structure is:
            - a/
              - env (export AAA="aaa")
              - b/
                - env (export BBB="bbb")
        and our current directory is `a/b`:
            >>> 'BBB' in os.environ
            <<< False
            >>>  with source('env'):
            ...     os.environ['BBB']
            <<< 'bbb'
            >>> 'BBB' in os.environ
            <<< False
            >>> source('env')
            >>> os.environ['BBB']
            <<< 'bbb'
            >>>  with source('env', recursive=True):
            ...     os.environ['AAA']
            <<< 'aaa'
            >>> 'AAA' in os.environ
            <<< False
            >>> source('env', recursive=True)
            >>> os.environ['AAA']
            <<< 'aa'
    """

    ptr = pathlib.Path('.').resolve()
    filenames = []
    if (ptr / filename).exists() and (ptr / filename).is_file():
        filenames.append(str((ptr / filename).resolve()))
    if recursive:
        while True:
            ptr = ptr.parent
            if (ptr / filename).exists() and (ptr / filename).is_file():
                filenames.append(str((ptr / filename).resolve()))
            if ptr == ptr.parent: break

    env = {}
    shell_cmd = globals()[shell]
    for filename in reversed(filenames):
        result = f"source {filename} && declare -x" | shell_cmd
        if not result:
            if quiet: continue
            else: result.raise_for_returncode()
        for line in result:
            match = re.search(r'^declare -x ([^=]+)="(.*)"$', line.strip())
            if not match: continue
            key, value = match.groups()
            if key not in os.environ or value != os.environ[key]: env[key] = value
    return export(**env)

